from unittest.mock import Mock

import pytest

import toga
from toga.sources import ListSource
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def on_select_handler():
    return Mock()


@pytest.fixture
def on_activate_handler():
    return Mock()


@pytest.fixture
def source():
    return ListSource(
        accessors=["key", "value"],
        data=[
            {"key": "first", "value": 111, "other": "aaa"},
            {"key": "second", "value": 222, "other": "bbb"},
            {"key": "third", "value": 333, "other": "ccc"},
        ],
    )


@pytest.fixture
def table(source, on_select_handler, on_activate_handler):
    return toga.Table(
        ["Title", "Value"],
        accessors=["key", "value"],
        data=source,
        on_select=on_select_handler,
        on_activate=on_activate_handler,
    )


def test_table_created():
    """A minimal Table can be created."""
    table = toga.Table(["First", "Second"])
    assert table._impl.interface == table
    assert_action_performed(table, "create Table")

    assert len(table.data) == 0
    assert table.headings == ["First", "Second"]
    assert table.accessors == ["first", "second"]
    assert not table.multiple_select
    assert table.missing_value == ""
    assert table.on_select._raw is None
    assert table.on_activate._raw is None


def test_create_with_values(source, on_select_handler, on_activate_handler):
    """A Table can be created with initial values."""
    table = toga.Table(
        ["First", "Second"],
        data=source,
        accessors=["primus", "secondus"],
        multiple_select=True,
        on_select=on_select_handler,
        on_activate=on_activate_handler,
        missing_value="Boo!",
    )
    assert table._impl.interface == table
    assert_action_performed(table, "create Table")

    assert len(table.data) == 3
    assert table.headings == ["First", "Second"]
    assert table.accessors == ["primus", "secondus"]
    assert table.multiple_select
    assert table.missing_value == "Boo!"
    assert table.on_select._raw == on_select_handler
    assert table.on_activate._raw == on_activate_handler


def test_create_with_accessor_overrides():
    """A Table can partially override accessors."""
    table = toga.Table(
        ["First", "Second"],
        accessors={"First": "override"},
    )
    assert table._impl.interface == table
    assert_action_performed(table, "create Table")

    assert len(table.data) == 0
    assert table.headings == ["First", "Second"]
    assert table.accessors == ["override", "second"]


def test_create_no_headings():
    """A Table can be created with no headings."""
    table = toga.Table(
        headings=None,
        accessors=["primus", "secondus"],
    )
    assert table._impl.interface == table
    assert_action_performed(table, "create Table")

    assert len(table.data) == 0
    assert table.headings is None
    assert table.accessors == ["primus", "secondus"]


def test_create_headings_required():
    """A Table requires either headings can be created with no headings."""
    with pytest.raises(
        ValueError,
        match=r"Cannot create a table without either headings or accessors",
    ):
        toga.Table()


def test_disable_no_op(table):
    """Table doesn't have a disabled state."""
    # Enabled by default
    assert table.enabled

    # Try to disable the widget
    table.enabled = False

    # Still enabled.
    assert table.enabled


def test_focus_noop(table):
    """Focus is a no-op."""

    table.focus()
    assert_action_not_performed(table, "focus")


@pytest.mark.parametrize(
    "data, all_attributes, extra_attributes",
    [
        # List of lists
        (
            [
                ["Alice", 123, "extra1"],
                ["Bob", 234, "extra2"],
                ["Charlie", 345, "extra3"],
            ],
            True,
            False,
        ),
        # List of tuples
        (
            [
                ("Alice", 123, "extra1"),
                ("Bob", 234, "extra2"),
                ("Charlie", 345, "extra3"),
            ],
            True,
            False,
        ),
        # List of dictionaries
        (
            [
                {"key": "Alice", "value": 123, "extra": "extra1"},
                {"key": "Bob", "value": 234, "extra": "extra2"},
                {"key": "Charlie", "value": 345, "extra": "extra3"},
            ],
            True,
            True,
        ),
        # List of bare data
        (
            [
                "Alice",
                1234,
                "Charlie",
            ],
            False,
            False,
        ),
    ],
)
def test_set_data(table, on_select_handler, data, all_attributes, extra_attributes):
    """Data can be set from a variety of sources."""

    # The selection hasn't changed yet.
    on_select_handler.assert_not_called()

    # Change the data
    table.data = data

    # This triggered the select handler
    on_select_handler.assert_called_once_with(table)

    # A ListSource has been constructed
    assert isinstance(table.data, ListSource)
    assert len(table.data) == 3

    # The accessors are mapped in order.
    assert table.data[0].key == "Alice"
    assert table.data[2].key == "Charlie"

    if all_attributes:
        assert table.data[1].key == "Bob"
        assert table.data[0].value == 123
        assert table.data[1].value == 234
        assert table.data[2].value == 345
    else:
        assert table.data[1].key == 1234

    if extra_attributes:
        assert table.data[0].extra == "extra1"
        assert table.data[1].extra == "extra2"
        assert table.data[2].extra == "extra3"


def test_single_selection(table, on_select_handler):
    """The current selection can be retrieved."""
    # Selection is initially empty
    assert table.selection is None
    on_select_handler.assert_not_called()

    # Select an item
    table._impl.simulate_selection(1)

    # Selection returns a single row
    assert table.selection == table.data[1]

    # Selection handler was triggered
    on_select_handler.assert_called_once_with(table)


def test_multiple_selection(source, on_select_handler):
    """A multi-select table can have the selection retrieved."""
    table = toga.Table(
        ["Title", "Value"],
        data=source,
        multiple_select=True,
        on_select=on_select_handler,
    )
    # Selection is initially empty
    assert table.selection == []
    on_select_handler.assert_not_called()

    # Select an item
    table._impl.simulate_selection([0, 2])

    # Selection returns a list of rows
    assert table.selection == [table.data[0], table.data[2]]

    # Selection handler was triggered
    on_select_handler.assert_called_once_with(table)


def test_activation(table, on_activate_handler):
    """A row can be activated."""

    # Activate an item
    table._impl.simulate_activate(1)

    # Activate handler was triggered; the activated row is provided
    on_activate_handler.assert_called_once_with(table, row=table.data[1])


def test_scroll_to_top(table):
    """A table can be scrolled to the top."""
    table.scroll_to_top()

    assert_action_performed_with(table, "scroll to row", row=0)


@pytest.mark.parametrize(
    "row, effective",
    [
        # Positive index
        (0, 0),
        (2, 2),
        # Greater index than available rows
        (10, 3),
        # Negative index
        (-1, 2),
        (-3, 0),
        # Greater negative index than available rows
        (-10, 0),
    ],
)
def test_scroll_to_row(table, row, effective):
    """A table can be scrolled to a specific row."""
    table.scroll_to_row(row)

    assert_action_performed_with(table, "scroll to row", row=effective)


def test_scroll_to_row_no_data(table):
    """If there's no data, scrolling is a no-op."""
    table.data.clear()

    table.scroll_to_row(5)

    assert_action_not_performed(table, "scroll to row")


def test_scroll_to_bottom(table):
    """A table can be scrolled to the top."""
    table.scroll_to_bottom()

    assert_action_performed_with(table, "scroll to row", row=2)


def test_insert_column_accessor(table):
    """A column can be inserted at an accessor."""
    table.insert_column("value", "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        table,
        "insert column",
        index=1,
        heading="New Column",
        accessor="extra",
    )
    assert table.headings == ["Title", "New Column", "Value"]
    assert table.accessors == ["key", "extra", "value"]


def test_insert_column_unknown_accessor(table):
    """If the insertion index accessor is unknown, an error is raised."""
    with pytest.raises(ValueError, match=r"'unknown' is not in list"):
        table.insert_column("unknown", "New Column", accessor="extra")


def test_insert_column_index(table):
    """A column can be inserted."""

    table.insert_column(1, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        table,
        "insert column",
        index=1,
        heading="New Column",
        accessor="extra",
    )
    assert table.headings == ["Title", "New Column", "Value"]
    assert table.accessors == ["key", "extra", "value"]


def test_insert_column_big_index(table):
    """A column can be inserted at an index bigger than the number of columns."""

    table.insert_column(100, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        table,
        "insert column",
        index=2,
        heading="New Column",
        accessor="extra",
    )
    assert table.headings == ["Title", "Value", "New Column"]
    assert table.accessors == ["key", "value", "extra"]


def test_insert_column_negative_index(table):
    """A column can be inserted at a negative index."""

    table.insert_column(-2, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        table,
        "insert column",
        index=0,
        heading="New Column",
        accessor="extra",
    )
    assert table.headings == ["New Column", "Title", "Value"]
    assert table.accessors == ["extra", "key", "value"]


def test_insert_column_big_negative_index(table):
    """A column can be inserted at a negative index larger than the number of
    columns."""

    table.insert_column(-100, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        table,
        "insert column",
        index=0,
        heading="New Column",
        accessor="extra",
    )
    assert table.headings == ["New Column", "Title", "Value"]
    assert table.accessors == ["extra", "key", "value"]


def test_insert_column_no_accessor(table):
    """A column can be inserted with a default accessor."""

    table.insert_column(1, "New Column")

    # The column was added
    assert_action_performed_with(
        table,
        "insert column",
        index=1,
        heading="New Column",
        accessor="new_column",
    )
    assert table.headings == ["Title", "New Column", "Value"]
    assert table.accessors == ["key", "new_column", "value"]


def test_insert_column_no_headings(source):
    """A column can be inserted into a table with no headings."""
    table = toga.Table(headings=None, accessors=["key", "value"], data=source)

    table.insert_column(1, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        table,
        "insert column",
        index=1,
        heading=None,
        accessor="extra",
    )
    assert table.headings is None
    assert table.accessors == ["key", "extra", "value"]


def test_insert_column_no_headings_missing_accessor(source):
    """An accessor is mandatory when adding a column to a table with no headings."""
    table = toga.Table(headings=None, accessors=["key", "value"], data=source)

    with pytest.raises(
        ValueError,
        match=r"Must specify an accessor on a table without headings",
    ):
        table.insert_column(1, "New Column")


def test_append_column(table):
    """A column can be appended."""
    table.append_column("New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        table,
        "insert column",
        index=2,
        heading="New Column",
        accessor="extra",
    )
    assert table.headings == ["Title", "Value", "New Column"]
    assert table.accessors == ["key", "value", "extra"]


def test_remove_column_accessor(table):
    """A column can be removed by accessor."""

    table.remove_column("value")

    # The column was removed
    assert_action_performed_with(
        table,
        "remove column",
        index=1,
    )
    assert table.headings == ["Title"]
    assert table.accessors == ["key"]


def test_remove_column_unknown_accessor(table):
    """If the column named for removal doesn't exist, an error is raised."""
    with pytest.raises(ValueError, match=r"'unknown' is not in list"):
        table.remove_column("unknown")


def test_remove_column_invalid_index(table):
    """If the index specified doesn't exist, an error is raised."""
    with pytest.raises(IndexError, match=r"list assignment index out of range"):
        table.remove_column(100)


def test_remove_column_index(table):
    """A column can be removed by index."""

    table.remove_column(1)

    # The column was removed
    assert_action_performed_with(
        table,
        "remove column",
        index=1,
    )
    assert table.headings == ["Title"]
    assert table.accessors == ["key"]


def test_remove_column_negative_index(table):
    """A column can be removed by index."""

    table.remove_column(-2)

    # The column was removed
    assert_action_performed_with(
        table,
        "remove column",
        index=0,
    )
    assert table.headings == ["Value"]
    assert table.accessors == ["value"]


def test_remove_column_no_headings(table):
    """A column can be removed when there are no headings."""
    table = toga.Table(
        headings=None,
        accessors=["primus", "secondus"],
    )

    table.remove_column(1)

    # The column was removed
    assert_action_performed_with(
        table,
        "remove column",
        index=1,
    )
    assert table.headings is None
    assert table.accessors == ["primus"]


def test_deprecated_names(on_activate_handler):
    """Deprecated names still work."""

    # Can't specify both on_double_click and on_activate
    with pytest.raises(
        ValueError,
        match=r"Cannot specify both on_double_click and on_activate",
    ):
        toga.Table(["First", "Second"], on_double_click=Mock(), on_activate=Mock())

    # on_double_click is redirected at construction
    with pytest.warns(
        DeprecationWarning,
        match="Table.on_double_click has been renamed Table.on_activate",
    ):
        table = toga.Table(["First", "Second"], on_double_click=on_activate_handler)

    # on_double_click accessor is redirected to on_activate
    with pytest.warns(
        DeprecationWarning,
        match="Table.on_double_click has been renamed Table.on_activate",
    ):
        assert table.on_double_click._raw == on_activate_handler

    assert table.on_activate._raw == on_activate_handler

    # on_double_click mutator is redirected to on_activate
    new_handler = Mock()
    with pytest.warns(
        DeprecationWarning,
        match="Table.on_double_click has been renamed Table.on_activate",
    ):
        table.on_double_click = new_handler

    assert table.on_activate._raw == new_handler

    # add_column redirects to insert
    table.add_column("New Column", "new_accessor")

    assert_action_performed_with(
        table,
        "insert column",
        index=2,
        heading="New Column",
        accessor="new_accessor",
    )
    assert table.headings == ["First", "Second", "New Column"]
    assert table.accessors == ["first", "second", "new_accessor"]



from __future__ import annotations

import re
import warnings
from decimal import ROUND_HALF_UP, Decimal, InvalidOperation

from toga.handlers import wrapped_handler

from .base import Widget

# Implementation notes
# ====================
#
# * `step`, `min` and `max` maintain an interface shadow copy of
#   their current values. This is because we use Decimal as a representation,
#   but all the implementations use floats. To ensure that we can round-trip
#   step/min/max values, we need to keep a local copy.
# * Decimal(3.7) yields "3.700000000...177". However, Decimal(str(3.7)) yields
#   "3.7". If the user provides a float, convert to a string first to ensure
#   that we don't introduce floating point error.


NUMERIC_RE = re.compile(r"[^0-9\.-]")


def _clean_decimal(value, step=None):
    # Decimal(3.7) yields "3.700000000...177".
    # However, Decimal(str(3.7)) yields "3.7". If the user provides a float,
    # convert to a string first.
    if isinstance(value, float):
        value = str(value)
    value = Decimal(value)

    if step is not None:
        # ROUND_DOWN would not be unreasonable for manually-entered strings, but it
        # interacts badly with native increment operations that use floats. For example,
        # 1.2 might be incremented to 1.299999997, and then rounded back down to 1.2,
        # so nothing would change.
        value = value.quantize(step, rounding=ROUND_HALF_UP)
    return value


def _clean_decimal_str(value):
    """Clean a string value"""
    # Replace any character that isn't a number, `.` or `-`
    value = NUMERIC_RE.sub("", value)
    # Remove any `-` not at the start of the string
    pos = value.find("-", 1)
    while pos != -1:
        value = value[:pos] + value[pos + 1 :]
        pos = value.find("-", pos + 1)

    # Only allow the first instance of `.`
    pos = value.find(".")
    if pos != -1:
        pos = value.find(".", pos + 1)
        while pos != -1:
            value = value[:pos] + value[pos + 1 :]
            pos = value.find(".", pos + 1)

    return value


class NumberInput(Widget):
    def __init__(
        self,
        id=None,
        style=None,
        step: Decimal = 1,
        min: Decimal | None = None,
        max: Decimal | None = None,
        value: Decimal | None = None,
        readonly: bool = False,
        on_change: callable | None = None,
        min_value: Decimal | None = None,  # DEPRECATED
        max_value: Decimal | None = None,  # DEPRECATED
    ):
        """Create a new number input widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param step: The amount that any increment/decrement operations will apply to
            the widget's current value.
        :param min: If provided, :any:`value` will be guaranteed to be greater than or
            equal to this minimum.
        :param max: If provided, :any:`value` will be guaranteed to be less than or
            equal to this maximum.
        :param value: The initial value for the widget.
        :param readonly: Can the value of the widget be modified by the user?
        :param on_change: A handler that will be invoked when the value of the
            widget changes.
        :param min_value: **DEPRECATED**; alias of ``min``.
        :param max_value: **DEPRECATED**; alias of ``max``.
        """
        super().__init__(id=id, style=style)

        ######################################################################
        # 2023-06: Backwards compatibility
        ######################################################################
        if min_value is not None:
            if min is not None:
                raise ValueError("Cannot specify both min and min_value")
            else:
                warnings.warn(
                    "NumberInput.min_value has been renamed NumberInput.min",
                    DeprecationWarning,
                )
                min = min_value
        if max_value is not None:
            if max is not None:
                raise ValueError("Cannot specify both max and max_value")
            else:
                warnings.warn(
                    "NumberInput.max_value has been renamed NumberInput.max",
                    DeprecationWarning,
                )
                max = max_value
        ######################################################################
        # End backwards compatibility
        ######################################################################

        # The initial setting of min requires calling get_value(),
        # which in turn interrogates min. Prime those values with
        # an empty starting value
        self._min = None
        self._max = None

        self.on_change = None
        self._impl = self.factory.NumberInput(interface=self)

        self.readonly = readonly
        self.step = step
        self.min = min
        self.max = max
        self.value = value

        self.on_change = on_change

    @property
    def readonly(self) -> bool:
        """Can the value of the widget be modified by the user?

        This only controls manual changes by the user (i.e., typing at the
        keyboard). Programmatic changes are permitted while the widget has
        ``readonly`` enabled.
        """
        return self._impl.get_readonly()

    @readonly.setter
    def readonly(self, value):
        self._impl.set_readonly(value)

    @property
    def step(self) -> Decimal:
        """The amount that any increment/decrement operations will apply to the
        widget's current value. (Not all backends provide increment and
        decrement buttons.)
        """
        return self._step

    @step.setter
    def step(self, step):
        try:
            self._step = _clean_decimal(step)
        except (ValueError, TypeError, InvalidOperation):
            raise ValueError("step must be a number")

        self._impl.set_step(self._step)

        # Re-assigning the min and max value forces the min/max to be requantized.
        self.min = self.min
        self.max = self.max

    @property
    def min(self) -> Decimal | None:
        """The minimum bound for the widget's value.

        Returns :any:`None` if there is no minimum bound.

        When setting this property, the current :attr:`value` and :attr:`max` will be
        clipped against the new minimum value.
        """
        return self._min

    @min.setter
    def min(self, new_min):
        try:
            new_min = _clean_decimal(new_min, self.step)

            # Clip widget's value to the new minimum
            if self.value is not None and self.value < new_min:
                self.value = new_min
        except (TypeError, ValueError, InvalidOperation):
            if new_min is None or new_min == "":
                new_min = None
            else:
                raise ValueError("min must be a number or None")

        # Clip the max value if it's inconsistent with the new min
        if self.max is not None and new_min is not None and new_min > self.max:
            self.max = new_min

        self._min = new_min
        self._impl.set_min_value(new_min)

    @property
    def max(self) -> Decimal | None:
        """The maximum bound for the widget's value.

        Returns :any:`None` if there is no maximum bound.

        When setting this property, the current :attr:`value` and :attr:`min` will be
        clipped against the new maximum value.
        """
        return self._max

    @max.setter
    def max(self, new_max):
        try:
            new_max = _clean_decimal(new_max, self.step)

            # Clip widget's value to the new maximum
            if self.value is not None and self.value > new_max:
                self.value = new_max
        except (TypeError, ValueError, InvalidOperation):
            if new_max is None or new_max == "":
                new_max = None
            else:
                raise ValueError("max must be a number or None")

        # Clip the min value if it's inconsistent with the new max
        if self.min is not None and new_max is not None and new_max < self.min:
            self.min = new_max

        self._max = new_max
        self._impl.set_max_value(new_max)

    @property
    def value(self) -> Decimal | None:
        """Current value of the widget, rounded to the same number of decimal
        places as :any:`step`, or ``None`` if no value has been set.

        If this property is set to a value outside of the min/max range, it will be
        clipped.

        While the widget is being edited by the user, it is possible for the UI
        to contain a value which is outside of the min/max range, or has too many
        decimal places. In this case, this property will return a value that has been
        clipped and rounded, and the visible text will be updated to match as soon as
        the widget loses focus.
        """
        # Get the value currently displayed by the widget. This *could*
        # be outside the min/max range.
        value = self._impl.get_value()

        # If the widget has a current value, clip it
        if value is not None:
            if self.min is not None and value < self.min:
                return self.min
            elif self.max is not None and value > self.max:
                return self.max
        return value

    @value.setter
    def value(self, value):
        try:
            value = _clean_decimal(value, self.step)

            if self.min is not None and value < self.min:
                value = self.min
            elif self.max is not None and value > self.max:
                value = self.max
        except (TypeError, ValueError, InvalidOperation):
            if value is None or value == "":
                value = None
            else:
                raise ValueError("value must be a number or None")

        self._impl.set_value(value)
        self.refresh()

    @property
    def on_change(self) -> callable:
        """The handler to invoke when the value of the widget changes."""
        return self._on_change

    @on_change.setter
    def on_change(self, handler):
        self._on_change = wrapped_handler(self, handler)

    ######################################################################
    # 2023-06: Backwards compatibility
    ######################################################################

    @property
    def min_value(self) -> Decimal | None:
        """**DEPRECATED**; alias of :attr:`min`."""
        warnings.warn(
            "NumberInput.min_value has been renamed NumberInput.min",
            DeprecationWarning,
        )
        return self.min

    @min_value.setter
    def min_value(self, value):
        warnings.warn(
            "NumberInput.min_value has been renamed NumberInput.min",
            DeprecationWarning,
        )
        self.min = value

    @property
    def max_value(self) -> Decimal | None:
        """**DEPRECATED**; alias of :attr:`max`."""
        warnings.warn(
            "NumberInput.max_value has been renamed NumberInput.max",
            DeprecationWarning,
        )
        return self.max

    @max_value.setter
    def max_value(self, value):
        warnings.warn(
            "NumberInput.max_value has been renamed NumberInput.max",
            DeprecationWarning,
        )
        self.max = value



import sys
from unittest.mock import Mock

import pytest

import toga_dummy

if sys.version_info >= (3, 10):
    from importlib.metadata import EntryPoint
else:
    # Before Python 3.10, entry_points did not support the group argument;
    # so, the backport package must be used on older versions.
    from importlib_metadata import EntryPoint

import toga.platform
from toga.platform import current_platform, get_current_platform, get_platform_factory


@pytest.fixture
def clean_env(monkeypatch):
    monkeypatch.delenv("TOGA_BACKEND")


@pytest.fixture
def platform_factory_1():
    return Mock()


@pytest.fixture
def platform_factory_2():
    return Mock()


def patch_platforms(monkeypatch, platforms):
    monkeypatch.setattr(
        sys,
        "modules",
        {f"{name}_module.factory": factory for name, factory, _ in platforms},
    )

    monkeypatch.setattr(
        toga.platform,
        "entry_points",
        Mock(
            return_value=[
                EntryPoint(
                    name=current_platform if is_current else name,
                    value=f"{name}_module",
                    group="self.backends",
                )
                for name, _, is_current in platforms
            ]
        ),
    )


def test_get_current_platform_desktop():
    assert (
        get_current_platform()
        == {
            "darwin": "macOS",
            "linux": "linux",
            "win32": "windows",
        }[sys.platform]
    )


def test_get_current_platform_android_inferred(monkeypatch):
    """Android platform can be inferred from existence of sys.getandroidapilevel."""
    monkeypatch.setattr(sys, "platform", "linux")
    try:
        # since there isn't an existing attribute of this name, it can't be patched.
        sys.getandroidapilevel = Mock(return_value=42)
        assert get_current_platform() == "android"
    finally:
        del sys.getandroidapilevel


def test_get_current_platform_android(monkeypatch):
    """Android platform can be obtained directly from sys.platform."""
    monkeypatch.setattr(sys, "platform", "android")
    try:
        # since there isn't an existing attribute of this name, it can't be patched.
        sys.getandroidapilevel = Mock(return_value=42)
        assert get_current_platform() == "android"
    finally:
        del sys.getandroidapilevel


def test_get_current_platform_iOS(monkeypatch):
    """IOS platform can be obtained directly from sys.platform."""
    monkeypatch.setattr(sys, "platform", "ios")
    assert get_current_platform() == "iOS"


def test_get_current_platform_web(monkeypatch):
    """Web platform can be obtained directly from sys.platform."""
    monkeypatch.setattr(sys, "platform", "emscripten")
    assert get_current_platform() == "web"


@pytest.mark.parametrize("value", ["freebsd12", "freebsd13", "freebsd14"])
def test_get_current_platform_freebsd(monkeypatch, value):
    """FreeBSD platform can be obtained directly from sys.platform."""
    monkeypatch.setattr(sys, "platform", value)
    assert get_current_platform() == "freeBSD"


def _get_platform_factory():
    get_platform_factory.cache_clear()
    factory = get_platform_factory()
    get_platform_factory.cache_clear()
    return factory


def test_no_platforms(monkeypatch, clean_env):
    patch_platforms(monkeypatch, [])
    with pytest.raises(
        RuntimeError,
        match=r"No Toga backend could be loaded.",
    ):
        _get_platform_factory()


def test_one_platform_installed(monkeypatch, clean_env):
    only_platform_factory = Mock()
    patch_platforms(monkeypatch, [("only_platform", only_platform_factory, False)])

    factory = _get_platform_factory()
    assert factory == only_platform_factory


def test_multiple_platforms_installed(monkeypatch, clean_env):
    current_platform_factory = Mock()
    other_platform_factory = Mock()
    patch_platforms(
        monkeypatch,
        [
            ("other_platform", other_platform_factory, False),
            ("current_platform", current_platform_factory, True),
        ],
    )

    factory = _get_platform_factory()
    assert factory == current_platform_factory


def test_multiple_platforms_installed_fail_both_appropriate(monkeypatch, clean_env):
    current_platform_factory_1 = Mock()
    current_platform_factory_2 = Mock()
    patch_platforms(
        monkeypatch,
        [
            ("current_platform_1", current_platform_factory_1, True),
            ("current_platform_2", current_platform_factory_2, True),
        ],
    )

    with pytest.raises(
        RuntimeError,
        match=(
            r"Multiple candidate toga backends found: \('current_platform_1_module' "
            r"\(.*\), 'current_platform_2_module' \(.*\)\). Uninstall the backends you "
            r"don't require, or use TOGA_BACKEND to specify a backend."
        ),
    ):
        _get_platform_factory()


def test_multiple_platforms_installed_fail_none_appropriate(monkeypatch, clean_env):
    other_platform_factory_1 = Mock()
    other_platform_factory_2 = Mock()
    patch_platforms(
        monkeypatch,
        [
            ("other_platform_1", other_platform_factory_1, False),
            ("other_platform_2", other_platform_factory_2, False),
        ],
    )

    with pytest.raises(
        RuntimeError,
        match=(
            r"Multiple Toga backends are installed \('other_platform_1_module' "
            r"\(.*\), 'other_platform_2_module' \(.*\)\), but none of them match "
            r"your current platform \('.*'\). Install a backend for your current "
            r"platform, or use TOGA_BACKEND to specify a backend."
        ),
    ):
        _get_platform_factory()


def test_environment_variable(monkeypatch):
    monkeypatch.setenv("TOGA_BACKEND", "toga_dummy")
    assert toga_dummy.factory == _get_platform_factory()


def test_environment_variable_fail(monkeypatch):
    monkeypatch.setenv("TOGA_BACKEND", "fake_platform_module")
    with pytest.raises(
        RuntimeError,
        match=r"The backend specified by TOGA_BACKEND \('fake_platform_module'\) could not be loaded.",
    ):
        _get_platform_factory()



from unittest.mock import MagicMock, call

import pytest

import toga
from toga_dummy.utils import (
    EventLog,
    assert_action_performed,
    attribute_value,
)


@pytest.fixture
def switch():
    return toga.Switch("Test Switch")


def test_widget_created(switch):
    """A switch can be created."""
    # Round trip the impl/interface
    assert switch._impl.interface == switch
    assert_action_performed(switch, "create Switch")

    assert switch.text == "Test Switch"
    assert not switch.value
    assert switch.on_change._raw is None
    assert switch.enabled


def test_widget_created_explicit(switch):
    """Explicit arguments at construction are stored."""

    def change_handler(widget, *args, **kwargs):
        pass

    switch = toga.Switch(
        "Explicit Switch",
        value=True,
        enabled=False,
        on_change=change_handler,
    )

    assert switch.text == "Explicit Switch"
    assert switch.on_change._raw == change_handler
    assert not switch.enabled


@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Text", "New Text"),
        ("", ""),
        (None, ""),
        ("\u200B", ""),
        (12345, "12345"),
        ("Contains\nnewline", "Contains"),
    ],
)
def test_label_text(switch, value, expected):
    """The switch's label can be modified."""
    assert switch.text == "Test Switch"

    # Clear the event log
    EventLog.reset()

    switch.text = value
    assert switch.text == expected

    # test backend has the right value
    assert attribute_value(switch, "text") == expected

    # A refresh was performed
    assert_action_performed(switch, "refresh")


@pytest.mark.parametrize(
    "value, expected",
    [
        ("A value", True),
        ("False", True),  # bool("False") - The literal string False evaluates as True
        ("", False),
        (None, False),
        (1234, True),
        (0, False),
    ],
)
def test_value_change(switch, value, expected):
    """The value of the switch can be set from almost any value."""
    switch.value = value
    assert switch.value == expected


def test_toggle(switch):
    """Toggle can be used to change the value."""
    assert not switch.value

    switch.toggle()
    assert switch.value

    switch.toggle()
    assert not switch.value


def test_on_change(switch):
    """The on_change handler is invoked whenever the value is changed."""
    handler = MagicMock()
    switch.on_change = handler

    # Reset the mock; installing the mock causes it to be evaluated as a bool()
    handler.reset_mock()

    # Set the value explicitly using a non-bool value
    switch.value = 100
    assert handler.mock_calls == [call(switch)]

    # Set the value explicitly using a boolean that is the same;
    # no signal as a result.
    switch.value = True
    assert handler.mock_calls == [call(switch)]

    # Toggle the switch
    switch.toggle()
    assert handler.mock_calls == [call(switch), call(switch)]



from __future__ import annotations

import sys
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import toga
from toga.platform import get_platform_factory

if TYPE_CHECKING:
    if sys.version_info < (3, 10):
        from typing_extensions import TypeAlias
    else:
        from typing import TypeAlias

    IconContent: TypeAlias = str | Path | toga.Icon


class cachedicon:
    def __init__(self, f):
        self.f = f
        self.__doc__ = f.__doc__

    def __get__(self, obj, owner):
        # If you ask for Icon.CACHED_ICON, obj is None, and owner is the Icon class
        # If you ask for self.CACHED_ICON, obj is self, from which we can get the class.
        if obj is None:
            cls = owner
        else:
            cls = obj.__class__

        try:
            # Look for a __CACHED_ICON attribute on the class
            value = getattr(cls, f"__{self.f.__name__}")
        except AttributeError:
            value = self.f(owner)
            setattr(cls, f"__{self.f.__name__}", value)
        return value


# A sentinel value that is type compatible with the `path` argument,
# but can be used to uniquely identify a request for an application icon
_APP_ICON = "<app icon>"


class Icon:
    @cachedicon
    def TOGA_ICON(cls) -> Icon:
        """**DEPRECATED** - Use ``DEFAULT_ICON``, or your own icon."""
        warnings.warn(
            "TOGA_ICON has been deprecated; Use DEFAULT_ICON, or your own icon.",
            DeprecationWarning,
        )

        return Icon("toga", system=True)

    @cachedicon
    def APP_ICON(cls) -> Icon:
        """The application icon.

        The application icon will be loaded from ``resources/<app name>`` (where ``<app
        name>`` is the value of :attr:`toga.App.app_name`).

        If this resource cannot be found, and the app has been packaged as a binary, the
        icon from the application binary will be used as a fallback.

        Otherwise, :attr:`~toga.Icon.DEFAULT_ICON` will be used.
        """
        return Icon(_APP_ICON)

    @cachedicon
    def DEFAULT_ICON(cls) -> Icon:
        """The default icon used as a fallback - Toga's "Tiberius the yak" icon."""
        return Icon("toga", system=True)

    @cachedicon
    def OPTION_CONTAINER_DEFAULT_TAB_ICON(cls) -> Icon:
        """The default icon used to decorate option container tabs."""
        return Icon("optioncontainer-tab", system=True)

    def __init__(
        self,
        path: str | Path,
        *,
        system: bool = False,  # Deliberately undocumented; for internal use only
    ):
        """Create a new icon.

        :param path: Base filename for the icon. The path can be an absolute file system
            path, or a path relative to the module that defines your Toga application
            class. This base filename should *not* contain an extension. If an extension
            is specified, it will be ignored. If the icon cannot be found, the default
            icon will be :attr:`~toga.Icon.DEFAULT_ICON`.
        :param system: **For internal use only**
        """
        self.factory = get_platform_factory()

        try:
            # Try to load the icon with the given path snippet. If the request is for the
            # app icon, use ``resources/<app name>`` as the path.
            if path is _APP_ICON:
                self.path = Path(f"resources/{toga.App.app.app_name}")
            else:
                self.path = Path(path)

            self.system = system
            if self.system:
                resource_path = Path(self.factory.__file__).parent / "resources"
            else:
                resource_path = toga.App.app.paths.app

            if self.factory.Icon.SIZES:
                full_path = {}
                for size in self.factory.Icon.SIZES:
                    try:
                        full_path[size] = self._full_path(
                            size=size,
                            extensions=self.factory.Icon.EXTENSIONS,
                            resource_path=resource_path,
                        )
                    except FileNotFoundError:
                        # This size variant wasn't found; we can skip it
                        pass
            else:
                full_path = self._full_path(
                    size=None,
                    extensions=self.factory.Icon.EXTENSIONS,
                    resource_path=resource_path,
                )

            self._impl = self.factory.Icon(interface=self, path=full_path)
        except FileNotFoundError:
            # Icon path couldn't be loaded. If the path is the sentinel for the app
            # icon, and this isn't running as a script, fall back to the application
            # binary
            if path is _APP_ICON:
                if Path(sys.executable).stem not in {
                    "python",
                    f"python{sys.version_info.major}",
                    f"python{sys.version_info.major}.{sys.version_info.minor}",
                }:
                    try:
                        # Use the application binary's icon
                        self._impl = self.factory.Icon(interface=self, path=None)
                    except FileNotFoundError:
                        # Can't find the application binary's icon.
                        print(
                            "WARNING: Can't find app icon; falling back to default icon"
                        )
                        self._impl = self.DEFAULT_ICON._impl
                else:
                    self._impl = self.DEFAULT_ICON._impl
            else:
                print(
                    f"WARNING: Can't find icon {self.path}; falling back to default icon"
                )
                self._impl = self.DEFAULT_ICON._impl

    def _full_path(self, size, extensions, resource_path):
        platform = toga.platform.current_platform
        if size:
            for extension in extensions:
                for filename in [
                    f"{self.path.stem}-{platform}-{size}{extension}",
                    f"{self.path.stem}-{size}{extension}",
                ]:
                    icon_path = resource_path / self.path.parent / filename
                    if icon_path.exists():
                        return icon_path

        # Look for size-less alternatives
        for extension in extensions:
            for filename in [
                f"{self.path.stem}-{platform}{extension}",
                f"{self.path.stem}{extension}",
            ]:
                icon_path = resource_path / self.path.parent / filename
                if icon_path.exists():
                    return icon_path

        raise FileNotFoundError(f"Can't find icon {self.path}")

    def __eq__(self, other):
        return isinstance(other, Icon) and other._impl.path == self._impl.path



from __future__ import annotations

from .base import Source


class ValueSource(Source):
    def __init__(self, value=None, accessor="value"):
        super().__init__()
        self.accessor = accessor
        setattr(self, accessor, value)

    def __str__(self):
        return str(getattr(self, self.accessor, None))

    def __setattr__(self, attr, value):
        super().__setattr__(attr, value)
        if attr == getattr(self, "accessor", None):
            self.notify("change", item=value)



from toga.style.pack import BOTTOM, COLUMN, LEFT, RIGHT, ROW, RTL, TOP, Pack

from ..utils import ExampleNode, ExampleViewport, assert_layout


def test_row_box_child_layout():
    root = ExampleNode(
        "container",
        style=Pack(text_direction=RTL, direction=ROW),
        children=[
            ExampleNode(
                "widget_a",
                style=Pack(width=30, height=100, padding_left=32, padding_right=34),
            ),
            ExampleNode(
                "widget_b",
                style=Pack(width=36, height=100, padding_left=38, padding_right=40),
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (210, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (576, 0), "content": (30, 100)},
                {"origin": (468, 0), "content": (36, 100)},
            ],
        },
    )


def test_column_box_child_layout():
    root = ExampleNode(
        "container",
        style=Pack(text_direction=RTL, direction=COLUMN),
        children=[
            ExampleNode(
                "widget_a",
                style=Pack(width=100, height=30, padding_top=32, padding_bottom=34),
            ),
            ExampleNode(
                "widget_b",
                style=Pack(width=100, height=36, padding_top=38, padding_bottom=40),
            ),
        ],
    )

    # Normal size
    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 210),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (0, 32), "content": (100, 30)},
                {"origin": (0, 134), "content": (100, 36)},
            ],
        },
    )


def test_alignment_top():
    root = ExampleNode(
        "root",
        style=Pack(text_direction=RTL, direction=ROW, alignment=TOP),
        children=[
            ExampleNode("space_filler", style=Pack(width=30, height=100)),
            ExampleNode("widget", style=Pack(width=30, height=30)),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (60, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (610, 0), "content": (30, 100)},
                {"origin": (580, 0), "content": (30, 30)},
            ],
        },
    )


def test_alignment_bottom():
    root = ExampleNode(
        "root",
        style=Pack(text_direction=RTL, direction=ROW, alignment=BOTTOM),
        children=[
            ExampleNode("space_filler", style=Pack(width=30, height=100)),
            ExampleNode("widget", style=Pack(width=30, height=30)),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (60, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (610, 380), "content": (30, 100)},
                {"origin": (580, 450), "content": (30, 30)},
            ],
        },
    )


def test_alignment_left():
    root = ExampleNode(
        "root",
        style=Pack(text_direction=RTL, direction=COLUMN, alignment=LEFT),
        children=[
            ExampleNode("space_filler", style=Pack(width=100, height=30)),
            ExampleNode("widget", style=Pack(width=30, height=30)),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 60),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (0, 0), "content": (100, 30)},
                {"origin": (0, 30), "content": (30, 30)},
            ],
        },
    )


def test_alignment_right():
    root = ExampleNode(
        "root",
        style=Pack(text_direction=RTL, direction=COLUMN, alignment=RIGHT),
        children=[
            ExampleNode("space_filler", style=Pack(width=100, height=30)),
            ExampleNode("widget", style=Pack(width=30, height=30)),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 60),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (540, 0), "content": (100, 30)},
                {"origin": (610, 30), "content": (30, 30)},
            ],
        },
    )



from datetime import date, datetime, time
from unittest.mock import Mock

import pytest

import toga
from toga_dummy.utils import assert_action_performed


@pytest.fixture
def on_change_handler():
    return Mock()


@pytest.fixture
def widget(on_change_handler):
    return toga.TimeInput(on_change=on_change_handler)


@pytest.mark.freeze_time("2023-05-25 10:42:37.123456")
def test_widget_created():
    """A TimeInput can be created."""
    widget = toga.TimeInput()

    # Round trip the impl/interface
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create TimeInput")

    assert widget.value == time(10, 42, 37)
    assert widget.on_change._raw is None


def test_widget_created_with_values(on_change_handler):
    """A TimeInput can be created with initial values."""
    # Round trip the impl/interface
    widget = toga.TimeInput(
        value=time(13, 37, 42),
        min=time(6, 1, 2),
        max=time(18, 58, 59),
        on_change=on_change_handler,
    )
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create TimeInput")

    assert widget.value == time(13, 37, 42)
    assert widget.min == time(6, 1, 2)
    assert widget.max == time(18, 58, 59)
    assert widget.on_change._raw == on_change_handler

    # The change handler isn't invoked at construction.
    on_change_handler.assert_not_called()


@pytest.mark.freeze_time("2023-05-25 10:42:37.123456")
@pytest.mark.parametrize(
    "value, expected",
    [
        (time(14, 37, 42, 123456), time(14, 37, 42)),
        (datetime(2023, 2, 11, 14, 37, 42, 123456), time(14, 37, 42)),
        ("14:37:42.123456", time(14, 37, 42)),
        (None, time(10, 42, 37)),
    ],
)
def test_value(widget, value, expected, on_change_handler):
    """The value of the datepicker can be set."""
    widget.value = value

    assert widget.value == expected

    on_change_handler.assert_called_once_with(widget)


INVALID_VALUES = [
    (123, TypeError, "Not a valid time value"),
    (object(), TypeError, "Not a valid time value"),
    (date(2023, 5, 25), TypeError, "Not a valid time value"),
    ("not a time", ValueError, "Invalid isoformat string: 'not a time'"),
]


@pytest.mark.parametrize("value, exc, message", INVALID_VALUES)
def test_invalid_value(widget, value, exc, message):
    """Invalid time values raise an exception."""
    with pytest.raises(exc, match=message):
        widget.value = value


@pytest.mark.parametrize(
    "value, clipped",
    [
        (time(3, 37, 42), time(6, 0, 0)),
        (time(5, 59, 59), time(6, 0, 0)),
        (time(6, 0, 0), time(6, 0, 0)),
        (time(10, 37, 42), time(10, 37, 42)),
        (time(18, 0, 0), time(18, 0, 0)),
        (time(18, 0, 1), time(18, 0, 0)),
        (time(22, 37, 42), time(18, 0, 0)),
    ],
)
def test_value_clipping(widget, value, clipped, on_change_handler):
    """It the value is inconsistent with min/max, it is clipped."""
    # Set min/max dates, and clear the on_change mock
    widget.min = time(6, 0, 0)
    widget.max = time(18, 0, 0)
    on_change_handler.reset_mock()

    # Set the new value
    widget.value = value

    # Value has been clipped
    assert widget.value == clipped

    # on_change handler called once.
    on_change_handler.assert_called_once_with(widget)


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, time(0, 0, 0)),
        (time(6, 1, 11), time(6, 1, 11)),
        (datetime(2023, 6, 16, 6, 2, 11), time(6, 2, 11)),
        ("06:03:11", time(6, 3, 11)),
    ],
)
def test_min(widget, value, expected):
    """The min of the datepicker can be set."""
    widget.min = value

    assert widget.min == expected


@pytest.mark.parametrize("value, exc, message", INVALID_VALUES)
def test_invalid_min(widget, value, exc, message):
    """Invalid min values raise an exception."""
    widget.max = time(18, 0, 0)

    with pytest.raises(exc, match=message):
        widget.min = value


@pytest.mark.parametrize(
    "min, clip_value, clip_max",
    [
        (time(1, 2, 3), False, False),
        (time(3, 42, 37), False, False),
        (time(3, 42, 38), True, False),
        (time(6, 0, 0), True, False),
        (time(12, 0, 0), True, False),
        (time(12, 0, 1), True, True),
        (time(13, 14, 15), True, True),
    ],
)
def test_min_clip(widget, on_change_handler, min, clip_value, clip_max):
    """If the current value or max is before a new min time, it is clipped."""
    widget.value = time(3, 42, 37)
    widget.max = time(12, 0, 0)
    on_change_handler.reset_mock()

    widget.min = min
    assert widget.min == min

    if clip_value:
        assert widget.value == min
        on_change_handler.assert_called_once_with(widget)
    else:
        assert widget.value == time(3, 42, 37)
        on_change_handler.assert_not_called()

    if clip_max:
        assert widget.max == min
    else:
        assert widget.max == time(12, 0, 0)


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, time(23, 59, 59)),
        (time(18, 1, 11), time(18, 1, 11)),
        (datetime(2023, 5, 25, 18, 2, 11), time(18, 2, 11)),
        ("18:03:11", time(18, 3, 11)),
    ],
)
def test_max(widget, value, expected):
    """The max of the datepicker can be set."""
    widget.max = value

    assert widget.max == expected


@pytest.mark.parametrize("value, exc, message", INVALID_VALUES)
def test_invalid_max(widget, value, exc, message):
    """Invalid max values raise an exception."""
    widget.min = time(18, 0, 0)

    with pytest.raises(exc, match=message):
        widget.max = value


@pytest.mark.parametrize(
    "max, clip_value, clip_min",
    [
        (time(1, 2, 3), True, True),
        (time(3, 42, 36), True, True),
        (time(3, 42, 37), True, False),
        (time(6, 0, 0), True, False),
        (time(11, 59, 0), True, False),
        (time(12, 0, 0), False, False),
        (time(13, 14, 15), False, False),
    ],
)
def test_max_clip(widget, on_change_handler, max, clip_value, clip_min):
    """If the current value is after a new max date, the value is clipped."""
    widget.min = time(3, 42, 37)
    widget.value = time(12, 0, 0)
    on_change_handler.reset_mock()

    widget.max = max
    assert widget.max == max

    if clip_value:
        assert widget.value == max
        on_change_handler.assert_called_once_with(widget)
    else:
        assert widget.value == time(12, 0, 0)
        on_change_handler.assert_not_called()

    if clip_min:
        assert widget.min == max
    else:
        assert widget.min == time(3, 42, 37)


def test_deprecated_names():
    MIN = time(8, 30, 59)
    MAX = time(10, 0, 0)

    with pytest.warns(
        DeprecationWarning, match="TimePicker has been renamed TimeInput"
    ), pytest.warns(
        DeprecationWarning, match="TimePicker.min_time has been renamed TimeInput.min"
    ), pytest.warns(
        DeprecationWarning, match="TimePicker.max_time has been renamed TimeInput.max"
    ):
        widget = toga.TimePicker(min_time=MIN, max_time=MAX)

    assert widget.min == MIN
    assert widget.max == MAX
    widget.min = widget.max = None

    with pytest.warns(
        DeprecationWarning, match="TimePicker.min_time has been renamed TimeInput.min"
    ):
        widget.min_time = MIN

    with pytest.warns(
        DeprecationWarning, match="TimePicker.min_time has been renamed TimeInput.min"
    ):
        assert widget.min_time == MIN
    assert widget.min == MIN

    with pytest.warns(
        DeprecationWarning, match="TimePicker.max_time has been renamed TimeInput.max"
    ):
        widget.max_time = MAX

    with pytest.warns(
        DeprecationWarning, match="TimePicker.max_time has been renamed TimeInput.max"
    ):
        assert widget.max_time == MAX
    assert widget.max == MAX

    with pytest.warns(
        DeprecationWarning, match="TimePicker has been renamed TimeInput"
    ):
        widget = toga.TimePicker()
    assert widget.min == time(0, 0, 0)
    assert widget.max == time(23, 59, 59)



from travertino.size import at_least

from toga.style.pack import COLUMN, ROW, Pack

from ..utils import ExampleNode, ExampleViewport, assert_layout


def test_tutorial_1():
    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN, padding_top=10),
        children=[
            ExampleNode(
                "f_box",
                style=Pack(direction=ROW, padding=5),
                children=[
                    ExampleNode(
                        "f_input",
                        style=Pack(flex=1, padding_left=160),
                        size=(at_least(100), 15),
                    ),
                    ExampleNode(
                        "f_label",
                        style=Pack(width=100, padding_left=10),
                        size=(at_least(40), 10),
                    ),
                ],
            ),
            ExampleNode(
                "c_box",
                style=Pack(direction=ROW, padding=5),
                children=[
                    ExampleNode(
                        "join_label",
                        style=Pack(width=150, padding_right=10),
                        size=(at_least(80), 10),
                    ),
                    ExampleNode(
                        "c_input", style=Pack(flex=1), size=(at_least(100), 15)
                    ),
                    ExampleNode(
                        "c_label",
                        style=Pack(width=100, padding_left=10),
                        size=(at_least(40), 10),
                    ),
                ],
            ),
            ExampleNode(
                "button", style=Pack(flex=1, padding=15), size=(at_least(120), 30)
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (380, 120),
        (640, 480),
        {
            "origin": (0, 10),
            "content": (640, 470),
            "children": [
                {
                    "origin": (5, 15),
                    "content": (630, 15),
                    "children": [
                        {"origin": (165, 15), "content": (360, 15)},
                        {"origin": (535, 15), "content": (100, 10)},
                    ],
                },
                {
                    "origin": (5, 40),
                    "content": (630, 15),
                    "children": [
                        {"origin": (5, 40), "content": (150, 10)},
                        {"origin": (165, 40), "content": (360, 15)},
                        {"origin": (535, 40), "content": (100, 10)},
                    ],
                },
                {"origin": (15, 75), "content": (610, 30)},
            ],
        },
    )



from __future__ import annotations

from typing import Protocol


class Listener(Protocol):
    """The protocol that must be implemented by objects that will act as a listener on a
    data source.
    """

    def change(self, item):
        """A change has occurred in an item.

        :param item: The data object that has changed.
        """

    def insert(self, index: int, item):
        """An item has been added to the data source.

        :param index: The 0-index position in the data.
        :param item: The data object that was added.
        """

    def remove(self, index: int, item):
        """An item has been removed from the data source.

        :param index: The 0-index position in the data.
        :param item: The data object that was added.
        """

    def clear(self):
        """All items have been removed from the data source."""


class Source:
    """A base class for data sources, providing an implementation of data notifications."""

    def __init__(self):
        self._listeners = []

    @property
    def listeners(self) -> list[Listener]:
        """The listeners of this data source.

        :returns: A list of objects that are listening to this data source.
        """
        return self._listeners

    def add_listener(self, listener: Listener):
        """Add a new listener to this data source.

        If the listener is already registered on this data source, the
        request to add is ignored.

        :param listener: The listener to add
        """
        if listener not in self._listeners:
            self._listeners.append(listener)

    def remove_listener(self, listener: Listener):
        """Remove a listener from this data source.

        :param listener: The listener to remove.
        """
        self._listeners.remove(listener)

    def notify(self, notification: str, **kwargs):
        """Notify all listeners an event has occurred.

        :param notification: The notification to emit.
        :param kwargs: The data associated with the notification.
        """
        for listener in self._listeners:
            try:
                method = getattr(listener, notification)
            except AttributeError:
                method = None

            if method:
                method(**kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

from toga.constants import FlashMode
from toga.handlers import AsyncResult, PermissionResult
from toga.platform import get_platform_factory

if TYPE_CHECKING:
    from toga.app import App


class PhotoResult(AsyncResult):
    RESULT_TYPE = "photo"


class CameraDevice:
    def __init__(self, impl):
        self._impl = impl

    @property
    def id(self) -> str:
        """A unique identifier for the device"""
        return self._impl.id()

    @property
    def name(self) -> str:
        """A human-readable name for the device"""
        return self._impl.name()

    @property
    def has_flash(self) -> bool:
        """Does the device have a flash?"""
        return self._impl.has_flash()

    def __eq__(self, other) -> bool:
        return self.id == other.id

    def __repr__(self) -> str:
        return f"<CameraDevice id={self.id} {self.name!r}>"

    def __str__(self) -> str:
        return self.name


class Camera:
    def __init__(self, app: App):
        self.factory = get_platform_factory()
        self._app = app
        self._impl = self.factory.Camera(self)

    @property
    def app(self) -> App:
        """The app with which the camera is associated"""
        return self._app

    @property
    def has_permission(self) -> bool:
        """Does the user have permission to use camera devices?

        If the platform requires the user to explicitly confirm permission, and
        the user has not yet given permission, this will return ``False``.
        """
        return self._impl.has_permission()

    def request_permission(self) -> PermissionResult:
        """Request sufficient permissions to capture photos.

        If permission has already been granted, this will return without prompting the
        user.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will start the process of requesting permissions, but will return
        *immediately*. The return value can be awaited in an asynchronous context, but
        cannot be used directly.

        :returns: An asynchronous result; when awaited, returns True if the app has
            permission to take a photo; False otherwise.
        """
        result = PermissionResult(None)

        if has_permission := self.has_permission:
            result.set_result(has_permission)
        else:
            self._impl.request_permission(result)

        return result

    @property
    def devices(self) -> list[CameraDevice]:
        """The list of available camera devices."""
        return [CameraDevice(impl) for impl in self._impl.get_devices()]

    def take_photo(
        self,
        device: CameraDevice | None = None,
        flash: FlashMode = FlashMode.AUTO,
    ) -> PhotoResult:
        """Capture a photo using one of the device's cameras.

        If the platform requires permission to access the camera, and the user hasn't
        previously provided that permission, this will cause permission to be requested.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will start the process of taking a photo, but will return
        *immediately*. The return value can be awaited in an asynchronous context, but
        cannot be used directly.

        :param device: The initial camera device to use. If a device is *not* specified,
            a default camera will be used. Depending on the hardware available, the user
            may be able to change the camera used to capture the image at runtime.
        :param flash: The initial flash mode to use; defaults to "auto". Depending on
            the hardware available, this may be modified by the user at runtime.
        :returns: An asynchronous result; when awaited, returns the :any:`toga.Image`
            captured by the camera, or ``None`` if the photo was  cancelled.
        :raises PermissionError: if the app does not have permission to use the camera.
        """
        photo = PhotoResult(None)
        self._impl.take_photo(photo, device=device, flash=flash)
        return photo



from unittest.mock import Mock

import pytest

from toga.sources import ListSource, Row


@pytest.fixture
def source():
    return ListSource(
        data=[
            {"val1": "first", "val2": 111},
            {"val1": "second", "val2": 222},
            {"val1": "third", "val2": 333},
        ],
        accessors=["val1", "val2"],
    )


@pytest.mark.parametrize(
    "value",
    [
        None,
        42,
        "not a list",
    ],
)
def test_invalid_accessors(value):
    """Accessors for a list source must be a list of attribute names."""
    with pytest.raises(
        ValueError,
        match=r"accessors should be a list of attribute names",
    ):
        ListSource(accessors=value)


def test_accessors_required():
    """A list source must specify *some* accessors."""
    with pytest.raises(
        ValueError,
        match=r"ListSource must be provided a list of accessors",
    ):
        ListSource(accessors=[], data=[1, 2, 3])


def test_accessors_copied():
    """A list source must specify *some* accessors."""
    accessors = ["foo", "bar"]
    source = ListSource(accessors)

    assert source._accessors == ["foo", "bar"]

    # The accessors have been copied.
    accessors.append("whiz")
    assert source._accessors == ["foo", "bar"]


def test_empty_source():
    """A list source can be constructed with no data."""
    source = ListSource(accessors=["foo", "bar"])

    assert len(source) == 0


def test_tuples():
    """A ListSource can be instantiated with tuples."""
    source = ListSource(
        data=[
            ("first", 111),
            ("second", 222),
            ("third", 333),
        ],
        accessors=["val1", "val2"],
    )

    assert len(source) == 3

    assert source[0].val1 == "first"
    assert source[0].val2 == 111

    assert source[1].val1 == "second"
    assert source[1].val2 == 222

    listener = Mock()
    source.add_listener(listener)

    # Set element 1
    source[1] = ("new element", 999)

    # source is the same size, but has different data
    assert len(source) == 3
    assert source[1].val1 == "new element"
    assert source[1].val2 == 999

    listener.insert.assert_called_once_with(index=1, item=source[1])


def test_list():
    """A ListSource can be instantiated with lists."""
    source = ListSource(
        data=[
            ["first", 111],
            ["second", 222],
            ["third", 333],
        ],
        accessors=["val1", "val2"],
    )

    assert len(source) == 3

    assert source[0].val1 == "first"
    assert source[0].val2 == 111

    assert source[1].val1 == "second"
    assert source[1].val2 == 222

    listener = Mock()
    source.add_listener(listener)

    # Set element 1
    source[1] = ["new element", 999]

    # source is the same size, but has different data
    assert len(source) == 3
    assert source[1].val1 == "new element"
    assert source[1].val2 == 999

    listener.insert.assert_called_once_with(index=1, item=source[1])


def test_dict():
    """A ListSource can be instantiated with dictionaries."""
    source = ListSource(
        data=[
            {"val1": "first", "val2": 111},
            {"val1": "second", "val2": 222},
            {"val1": "third", "val2": 333},
        ],
        accessors=["val1", "val2"],
    )

    assert len(source) == 3

    assert source[0].val1 == "first"
    assert source[0].val2 == 111

    assert source[1].val1 == "second"
    assert source[1].val2 == 222

    listener = Mock()
    source.add_listener(listener)

    # Set element 1
    source[1] = ["new element", 999]

    # source is the same size, but has different data
    assert len(source) == 3
    assert source[1].val1 == "new element"
    assert source[1].val2 == 999

    listener.insert.assert_called_once_with(index=1, item=source[1])


def test_flat_list():
    """A list source can be created from a flat list of objects."""

    class MyObject:
        def __init__(self, info):
            self.info = info

        def __str__(self):
            return "string value %s" % self.info

    data = [
        MyObject(True),
        MyObject(2),
        MyObject("string"),
    ]

    source = ListSource(
        data=data,
        accessors=["col1"],
    )

    for i, row in enumerate(source):
        assert row.col1 == data[i]


def test_flat_list_numbers():
    """A list source can be created from a flat list of numbers."""

    data = [
        100,
        200.0,
        -3.14,
    ]

    source = ListSource(
        data=data,
        accessors=["col1"],
    )

    for i, row in enumerate(source):
        assert row.col1 == data[i]


def test_flat_list_strings():
    """A list source can be created from a flat list of numbers."""

    data = [
        "xxx",
        "yyy",
        "zzz",
    ]

    source = ListSource(
        data=data,
        accessors=["col1"],
    )

    for i, row in enumerate(source):
        assert row.col1 == data[i]


def test_iter(source):
    """A list source can be iterated over."""
    result = 0
    for row in source:
        result += row.val2

    assert result == 666


def test_clear(source):
    """A list source can be cleared."""

    assert len(source) == 3

    listener = Mock()
    source.add_listener(listener)

    # Clear the list
    source.clear()

    # List is empty
    assert len(source) == 0

    # A notification was sent
    listener.clear.assert_called_once_with()


def test_insert_kwarg(source):
    """You can insert into a list source using kwargs."""

    listener = Mock()
    source.add_listener(listener)

    # Insert the new element
    row = source.insert(1, dict(val1="new element", val2=999))

    assert len(source) == 4
    assert source[1] == row
    assert row.val1 == "new element"
    assert row.val2 == 999

    listener.insert.assert_called_once_with(index=1, item=row)


def test_insert_positional(source):
    """You can insert into a list source using positional args."""
    listener = Mock()
    source.add_listener(listener)

    # Insert the new element using positional args.
    # The values are mapped to the accessors in order.
    row = source.insert(1, ("new element", 999))

    assert len(source) == 4
    assert source[1] == row
    assert row.val1 == "new element"
    assert row.val2 == 999

    listener.insert.assert_called_once_with(index=1, item=row)


def test_append_dict(source):
    """You can append onto a list source using a dictionary."""

    listener = Mock()
    source.add_listener(listener)

    # Append the new element
    row = source.append(dict(val1="new element", val2=999))

    assert len(source) == 4
    assert source[3] == row
    assert row.val1 == "new element"
    assert row.val2 == 999

    listener.insert.assert_called_once_with(index=3, item=row)


def test_append_positional(source):
    """You can append onto a list source using positional args."""
    listener = Mock()
    source.add_listener(listener)

    # Append the new element using positional args.
    # The values are mapped to the accessors in order.
    row = source.append(("new element", 999))

    assert len(source) == 4
    assert source[3] == row
    assert row.val1 == "new element"
    assert row.val2 == 999

    listener.insert.assert_called_once_with(index=3, item=row)


def test_del(source):
    """You can delete an item from a list source by index."""
    listener = Mock()
    source.add_listener(listener)

    # Delete the second element
    row = source[1]
    del source[1]

    assert len(source) == 2
    assert source[0].val1 == "first"
    assert source[0].val2 == 111

    assert source[1].val1 == "third"
    assert source[1].val2 == 333

    listener.remove.assert_called_once_with(item=row, index=1)


def test_remove(source):
    """You can remove an item from a list source."""
    listener = Mock()
    source.add_listener(listener)

    # Remove the second element
    row = source[1]
    source.remove(row)

    assert len(source) == 2
    assert source[0].val1 == "first"
    assert source[0].val2 == 111

    assert source[1].val1 == "third"
    assert source[1].val2 == 333

    listener.remove.assert_called_once_with(item=row, index=1)


def test_index(source):
    """You can get the index of any row within a list source."""
    for i, row in enumerate(source):
        assert i == source.index(row)

    # look-alike rows are not equal, so index lookup should fail
    lookalike_row = Row(val1="second", val2=222)
    with pytest.raises(
        ValueError,
        match=r"<Row .* val1='second' val2=222> is not in list",
    ):
        source.index(lookalike_row)

    with pytest.raises(
        ValueError,
        match=r"None is not in list",
    ):
        source.index(None)

    with pytest.raises(
        ValueError,
        match=r"<Row .* \(no attributes\)> is not in list",
    ):
        source.index(Row())


def test_find(source):
    """You can find the index of any matching row within a list source."""

    # Duplicate row 1 of the data.
    source.append(dict(val1="second", val2=222))

    # A unique row can be found
    assert source.find(dict(val1="third", val2=333)) == source[2]

    # A unique row can be found, using implied accessor order
    assert source.find(("third", 333)) == source[2]

    # A unique row can be found, using only the first accessor
    assert source.find("third") == source[2]

    # If data isn't unique, the first match is returned
    assert source.find(dict(val1="second", val2=222)) == source[1]

    # The search can start after a given instance
    assert source.find(dict(val1="second", val2=222), start=source[1]) == source[3]

    # A partial match is enough
    assert source.find(dict(val1="third")) == source[2]

    # find will fail if the object doesn't exist
    with pytest.raises(
        ValueError,
        match=r"No row matching {'val1': 'not there', 'val2': 999} in data",
    ):
        source.find(dict(val1="not there", val2=999))

    # An overspecified search will fail
    with pytest.raises(
        ValueError,
        match=r"No row matching {'val1': 'first', 'val2': 111, 'value': 'overspecified'} in data",
    ):
        source.find(dict(val1="first", val2=111, value="overspecified"))



import toga
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


def test_create_box():
    """A Box can be created."""
    box = toga.Box()
    # Round trip the impl/interface
    assert box._impl.interface == box

    assert_action_performed(box, "create Box")
    assert_action_not_performed(box, "add child")


def test_create_box_with_children():
    """A Box can be created with children."""
    child1 = toga.Box()
    child2 = toga.Box()
    box = toga.Box(children=[child1, child2])

    # Round trip the impl/interface
    assert box._impl.interface == box

    assert_action_performed(box, "create Box")
    assert_action_performed_with(box, "add child", child=child1._impl)
    assert_action_performed_with(box, "add child", child=child2._impl)

    # But the box will have children.
    assert box.children == [child1, child2]


def test_disable_no_op():
    """Box doesn't have a disabled state."""
    box = toga.Box()

    # Enabled by default
    assert box.enabled

    # Try to disable the widget
    box.enabled = False

    # Still enabled.
    assert box.enabled


def test_focus_noop():
    """Focus is a no-op."""
    box = toga.Box()

    box.focus()
    assert_action_not_performed(box, "focus")



from enum import Enum, auto

from travertino.constants import *  # noqa: F401, F403  pragma: no cover


class Direction(Enum):
    "The direction a given property should act"
    HORIZONTAL = 0
    VERTICAL = 1


class Baseline(Enum):
    "The meaning of a Y coordinate when drawing text."
    ALPHABETIC = auto()  #: Alphabetic baseline of the first line
    TOP = auto()  #: Top of text
    MIDDLE = auto()  #: Middle of text
    BOTTOM = auto()  #: Bottom of text


class FillRule(Enum):
    "The rule to use when filling paths."
    EVENODD = 0
    NONZERO = 1


##########################################################################
# Camera
##########################################################################


class FlashMode(Enum):
    """The flash mode to use when capturing photos or videos."""

    # These constant values allow `flash=True` and `flash=False` to work
    AUTO = -1
    OFF = 0
    ON = 1

    def __str__(self):
        return self.name.title()


# class VideoQuality(Enum):
#     """The quality of the video recording.
#
#     The values of ``LOW``, ``MEDIUM`` and ``HIGH`` represent specific (platform
#     dependent) resolutions. These resolutions will remain the same over time.
#
#     The values of ``CELLULAR`` and ``WIFI`` may change over time to reflect the
#     capabilities of network hardware.
#
#     ``HIGHEST`` will always refer to the highest quality that the device can
#     record.
#     """
#
#     LOW = 0
#     MEDIUM = 1
#     HIGH = 2
#
#     # Qualitative alternatives to these constants
#     CELLULAR = 0
#     WIFI = 1
#     HIGHEST = 2



from __future__ import annotations

from toga.handlers import wrapped_handler

from .base import Widget


class Switch(Widget):
    def __init__(
        self,
        text,
        id=None,
        style=None,
        on_change: callable | None = None,
        value: bool = False,
        enabled: bool = True,
    ):
        """Create a new Switch widget.

        :param text: The text to display beside the switch.
        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param value: The initial value for the switch.
        :param on_change: A handler that will be invoked when the switch changes
            value.
        :param enabled: Is the switch enabled (i.e., can it be pressed?).
            Optional; by default, switches are created in an enabled state.
        """
        super().__init__(id=id, style=style)

        self._impl = self.factory.Switch(interface=self)

        self.text = text

        # Set a dummy handler before installing the actual on_change, because we do not want
        # on_change triggered by the initial value being set
        self.on_change = None
        self.value = value

        self.on_change = on_change

        self.enabled = enabled

    @property
    def text(self) -> str:
        """The text label for the Switch.

        ``None``, and the Unicode codepoint U+200B (ZERO WIDTH SPACE), will be
        interpreted and returned as an empty string. Any other object will be
        converted to a string using ``str()``.

        Only one line of text can be displayed. Any content after the first
        newline will be ignored.
        """
        return self._impl.get_text()

    @text.setter
    def text(self, value):
        if value is None or value == "\u200B":
            value = ""
        else:
            # Switch text can't include line breaks. Strip any content
            # after a line break (if provided)
            value = str(value).split("\n")[0]

        self._impl.set_text(value)
        self.refresh()

    @property
    def on_change(self) -> callable:
        """The handler to invoke when the value of the switch changes."""
        return self._on_change

    @on_change.setter
    def on_change(self, handler):
        self._on_change = wrapped_handler(self, handler)

    @property
    def value(self) -> bool:
        """The current state of the switch, as a Boolean.

        Any non-Boolean value will be converted to a Boolean.
        """
        return self._impl.get_value()

    @value.setter
    def value(self, value):
        self._impl.set_value(bool(value))

    def toggle(self):
        """Reverse the current value of the switch."""
        self.value = not self.value



from __future__ import annotations

from typing import Any

from .base import Source
from .list_source import Row, _find_item


class Node(Row):
    def __init__(self, **data):
        """Create a new Node object.

        The keyword arguments specified in the constructor will be converted into
        attributes on the new object.

        When initially constructed, the Node will be a leaf node (i.e., no children,
        and marked unable to have children).

        When any public attributes of the node are modified (i.e., any attribute whose
        name doesn't start with ``_``), the source to which the node belongs will be
        notified.
        """
        super().__init__(**data)
        self._children: list[Node] | None = None
        self._parent: Node | None = None

    def __repr__(self):
        descriptor = " ".join(
            f"{attr}={getattr(self, attr)!r}"
            for attr in sorted(self.__dict__)
            if not attr.startswith("_")
        )
        if not descriptor:
            descriptor = "(no attributes)"
        if self._children is not None:
            descriptor += f"; {len(self)} children"

        return f"<{'Leaf ' if self._children is None else ''}Node {id(self):x} {descriptor}>"

    ######################################################################
    # Methods required by the TreeSource interface
    ######################################################################

    def __getitem__(self, index: int) -> Node:
        if self._children is None:
            raise ValueError(f"{self} is a leaf node")

        return self._children[index]

    def __delitem__(self, index: int):
        if self._children is None:
            raise ValueError(f"{self} is a leaf node")

        child = self._children[index]
        del self._children[index]

        # Child isn't part of this source, or a child of this node anymore.
        child._parent = None
        child._source = None

        self._source.notify("remove", parent=self, index=index, item=child)

    def __len__(self) -> int:
        if self.can_have_children():
            return len(self._children)
        else:
            return 0

    def can_have_children(self) -> bool:
        """Can the node have children?

        A value of :any:`True` does not necessarily mean the node *has* any children,
        only that the node is *allowed* to have children. The value of ``len()`` for
        the node indicates the number of actual children.
        """
        return self._children is not None

    ######################################################################
    # Utility methods to make TreeSource more list-like
    ######################################################################

    def __iter__(self):
        return iter(self._children or [])

    def __setitem__(self, index: int, data: Any):
        """Set the value of a specific child in the Node.

        :param index: The index of the child to change
        :param data: The data for the updated child. This data will be converted
            into a Node object.
        """
        if self._children is None:
            raise ValueError(f"{self} is a leaf node")

        old_node = self._children[index]
        old_node._parent = None
        old_node._source = None

        node = self._source._create_node(parent=self, data=data)
        self._children[index] = node
        self._source.notify("change", item=node)

    def insert(self, index: int, data: Any, children: Any = None):
        """Insert a node as a child of this node a specific index.

        :param index: The index at which to insert the new child.
        :param data: The data to insert into the Node as a child. This data will be
            converted into a Node object.
        :param children: The data for the children of the new child node.
        :returns: The new added child Node object.
        """
        if self._children is None:
            self._children = []

        if index < 0:
            index = max(len(self) + index, 0)
        else:
            index = min(len(self), index)

        node = self._source._create_node(parent=self, data=data, children=children)
        self._children.insert(index, node)
        self._source.notify("insert", parent=self, index=index, item=node)
        return node

    def append(self, data: Any, children: Any = None):
        """Append a node to the end of the list of children of this node.

        :param data: The data to append as a child of this node. This data will be
            converted into a Node object.
        :param children: The data for the children of the new child node.
        :returns: The new added child Node object.
        """
        return self.insert(len(self), data=data, children=children)

    def remove(self, child: Node):
        """Remove a child node from this node.

        :param child: The child node to remove from this node.
        """
        # Index will raise ValueError if the node is a leaf
        del self[self.index(child)]

    def index(self, child: Node):
        """The index of a specific node in children of this node.

        This search uses Node instances, and searches for an *instance* match.
        If two Node instances have the same values, only the Node that is the
        same Python instance will match. To search for values based on equality,
        use :meth:`~toga.sources.Node.find`.

        :param child: The node to find in the children of this node.
        :returns: The index of the node in the children of this node.
        :raises ValueError: If the node cannot be found in children of this node.
        """
        if self._children is None:
            raise ValueError(f"{self} is a leaf node")

        return self._children.index(child)

    def find(self, data: Any, start: Node = None):
        """Find the first item in the child nodes of this node that matches all the
        provided attributes.

        This is a value based search, rather than an instance search. If two Node
        instances have the same values, the first instance that matches will be
        returned. To search for a second instance, provide the first found instance as
        the ``start`` argument. To search for a specific Node instance, use the
        :meth:`~toga.sources.Node.index`.

        :param data: The data to search for. Only the values specified in data will be
            used as matching criteria; if the node contains additional data attributes,
            they won't be considered as part of the match.
        :param start: The instance from which to start the search. Defaults to ``None``,
            indicating that the first match should be returned.
        :return: The matching Node object.
        :raises ValueError: If no match is found.
        :raises ValueError: If the node is a leaf node.
        """
        if self._children is None:
            raise ValueError(f"{self} is a leaf node")

        return _find_item(
            candidates=self._children,
            data=data,
            accessors=self._source._accessors,
            start=start,
            error=f"No child matching {data!r} in {self}",
        )


class TreeSource(Source):
    def __init__(self, accessors: list[str], data: dict | list[tuple] | None = None):
        super().__init__()
        if isinstance(accessors, str) or not hasattr(accessors, "__iter__"):
            raise ValueError("accessors should be a list of attribute names")

        # Copy the list of accessors
        self._accessors = [a for a in accessors]
        if len(self._accessors) == 0:
            raise ValueError("TreeSource must be provided a list of accessors")

        if data:
            self._roots = self._create_nodes(parent=None, value=data)
        else:
            self._roots = []

    ######################################################################
    # Methods required by the TreeSource interface
    ######################################################################

    def __len__(self) -> int:
        return len(self._roots)

    def __getitem__(self, index: int) -> Node:
        return self._roots[index]

    def __delitem__(self, index: int):
        node = self._roots[index]
        del self._roots[index]
        node._source = None
        self.notify("remove", parent=None, index=index, item=node)

    ######################################################################
    # Factory methods for new nodes
    ######################################################################

    def _create_node(
        self,
        parent: Node | None,
        data: Any,
        children: list | dict | None = None,
    ):
        if isinstance(data, dict):
            node = Node(**data)
        elif hasattr(data, "__iter__") and not isinstance(data, str):
            node = Node(**dict(zip(self._accessors, data)))
        else:
            node = Node(**{self._accessors[0]: data})

        node._parent = parent
        node._source = self

        if children is not None:
            node._children = self._create_nodes(parent=node, value=children)

        return node

    def _create_nodes(self, parent: Node | None, value: Any):
        if isinstance(value, dict):
            return [
                self._create_node(parent=parent, data=data, children=children)
                for data, children in value.items()
            ]
        elif hasattr(value, "__iter__") and not isinstance(value, str):
            return [
                self._create_node(parent=parent, data=item[0], children=item[1])
                for item in value
            ]
        else:
            return [self._create_node(parent=parent, data=value)]

    ######################################################################
    # Utility methods to make TreeSources more list-like
    ######################################################################

    def __setitem__(self, index: int, data: Any):
        """Set the value of a specific root item in the data source.

        :param index: The root item to change
        :param data: The data for the updated item. This data will be converted
            into a Node object.
        """
        old_root = self._roots[index]
        old_root._parent = None
        old_root._source = None

        root = self._create_node(parent=None, data=data)
        self._roots[index] = root
        self.notify("change", item=root)

    def clear(self):
        """Clear all data from the data source."""
        self._roots = []
        self.notify("clear")

    def insert(
        self,
        index: int,
        data: Any,
        children: Any = None,
    ):
        """Insert a root node into the data source at a specific index.

        If the node is a leaf node, it will be converted into a non-leaf node.

        :param index: The index into the list of children at which to insert the item.
        :param data: The data to insert into the TreeSource. This data will be converted
            into a Node object.
        :param children: The data for the children to insert into the TreeSource.
        :returns: The newly constructed Node object.
        :raises ValueError: If the provided parent is not part of this TreeSource.
        """
        if index < 0:
            index = max(len(self) + index, 0)
        else:
            index = min(len(self), index)

        node = self._create_node(parent=None, data=data, children=children)
        self._roots.insert(index, node)
        node._parent = None
        self.notify("insert", parent=None, index=index, item=node)
        return node

    def append(self, data: Any, children: Any = None):
        """Append a root node at the end of the list of children of this source.

        If the node is a leaf node, it will be converted into a non-leaf node.

        :param data: The data to append onto the list of children of the given parent.
            This data will be converted into a Node object.
        :param children: The data for the children to insert into the TreeSource.
        :returns: The newly constructed Node object.
        :raises ValueError: If the provided parent is not part of this TreeSource.
        """
        return self.insert(len(self), data=data, children=children)

    def remove(self, node: Node):
        """Remove a node from the data source.

        This will also remove the node if it is a descendant of a root node.

        :param node: The node to remove from the data source.
        """
        if node._source != self:
            raise ValueError(f"{node} is not managed by this data source")

        if node._parent is None:
            del self[self.index(node)]
        else:
            node._parent.remove(node)

    def index(self, node: Node) -> int:
        """The index of a specific root node in the data source.

        This search uses Node instances, and searches for an *instance* match.
        If two Node instances have the same values, only the Node that is the
        same Python instance will match. To search for values based on equality,
        use :meth:`~toga.sources.TreeSource.find`.

        :param node: The node to find in the data source.
        :returns: The index of the node in the child list it is a part of.
        :raises ValueError: If the node cannot be found in the data source.
        """
        return self._roots.index(node)

    def find(self, data: Any, start: Node = None):
        """Find the first item in the child nodes of the given node that matches all the
        provided attributes.

        This is a value based search, rather than an instance search. If two Node
        instances have the same values, the first instance that matches will be
        returned. To search for a second instance, provide the first found instance as
        the ``start`` argument. To search for a specific Node instance, use the
        :meth:`~toga.sources.TreeSource.index`.

        :param data: The data to search for. Only the values specified in data will be
            used as matching criteria; if the node contains additional data attributes,
            they won't be considered as part of the match.
        :param start: The instance from which to start the search. Defaults to ``None``,
            indicating that the first match should be returned.
        :return: The matching Node object.
        :raises ValueError: If no match is found.
        :raises ValueError: If the provided parent is not part of this TreeSource.
        """
        return _find_item(
            candidates=self._roots,
            data=data,
            accessors=self._accessors,
            start=start,
            error=f"No root node matching {data!r} in {self}",
        )



import contextlib
from unittest.mock import Mock

import pytest

import toga
from toga_dummy import factory
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
)


def test_no_location(monkeypatch, app):
    """If there's no location service, and no factory implementation, accessing camera raises an
    exception."""
    try:
        monkeypatch.delattr(app, "_location")
    except AttributeError:
        pass
    monkeypatch.delattr(factory, "Location")

    # Accessing the location object should raise NotImplementedError
    with pytest.raises(NotImplementedError):
        app.location


@pytest.mark.parametrize(
    "initial, should_request, has_permission",
    [
        (-2, True, True),
        (-1, True, True),
        (0, True, False),
        (1, False, True),
        (2, False, True),
    ],
)
def test_request_permission(app, initial, should_request, has_permission):
    """An app can request permission to use location."""
    # The location instance round-trips the app instance
    assert app.location.app == app

    # Set initial permission
    app.location._impl._has_permission = initial

    assert (
        app.loop.run_until_complete(app.location.request_permission()) == has_permission
    )

    if should_request:
        assert_action_performed(app.location, "request permission")
        assert_action_not_performed(app.location, "request background permission")
    else:
        assert_action_not_performed(app.location, "request permission")
        assert_action_not_performed(app.location, "request background permission")

    # As a result of requesting, location permission is as expected
    assert app.location.has_permission == has_permission


def test_request_permission_sync(app):
    """An app can synchronously request permission to use location."""
    # Set initial permission
    app.location._impl._has_permission = -1

    result = app.location.request_permission()

    # This will cause a permission request to occur...
    assert_action_performed(app.location, "request permission")
    assert_action_not_performed(app.location, "request background permission")

    # ... but the result won't be directly comparable
    with pytest.raises(RuntimeError):
        # == True isn't good python, but it's going to raise an exception anyway.
        result == True  # noqa: E712


@pytest.mark.parametrize(
    "foreground, initial, raise_error, should_request, has_background_permission",
    [
        (-1, -1, True, False, False),
        (0, -1, True, False, False),
        (1, -1, False, True, True),
        (-1, 0, True, False, False),
        (0, 0, True, False, False),
        (1, 0, False, True, False),
        # -1, 1 can't happen; background can't be approved if foreground isn't confirmed
        # 0, 1 can't happen; background can't be approved if foreground was rejected
        (1, 1, False, False, True),
    ],
)
def test_request_background_permission(
    app, foreground, initial, raise_error, should_request, has_background_permission
):
    """An app can request background permission to use location."""
    # The location instance round-trips the app instance
    assert app.location.app == app

    # Set initial permissions
    app.location._impl._has_permission = foreground
    app.location._impl._has_background_permission = initial

    if raise_error:
        error_context = pytest.raises(
            PermissionError,
            match=(
                r"Cannot ask for background location permission "
                r"before confirming foreground location permission\."
            ),
        )
    else:
        error_context = contextlib.nullcontext()

    with error_context:
        assert (
            app.loop.run_until_complete(app.location.request_background_permission())
            == has_background_permission
        )

    if should_request:
        assert_action_not_performed(app.location, "request permission")
        assert_action_performed(app.location, "request background permission")
    else:
        assert_action_not_performed(app.location, "request permission")
        assert_action_not_performed(app.location, "request background permission")

    # As a result of requesting, location permission is as expected
    assert app.location.has_background_permission == has_background_permission


def test_request_background_permission_sync(app):
    """An app can synchronously request background permission to use location."""
    # Set initial permission
    app.location._impl._has_permission = 1
    app.location._impl._has_background_permission = -1

    result = app.location.request_background_permission()

    # This will cause a permission request to occur...
    assert_action_not_performed(app.location, "request permission")
    assert_action_performed(app.location, "request background permission")

    # ... but the result won't be directly comparable
    with pytest.raises(RuntimeError):
        # == True isn't good python, but it's going to raise an exception anyway.
        result == True  # noqa: E712


def test_current_location_prior_permission(app):
    """If permission has been previously requested, the current location can be determined."""
    # Set permission
    app.location._impl._has_permission = 1

    result = app.loop.run_until_complete(app.location.current_location())

    # Location was returned
    assert result == toga.LatLng(13, 25)

    assert_action_performed(app.location, "has permission")
    assert_action_performed(app.location, "get next location")

    # Install an on_change handler
    change_handler = Mock()
    app.location.on_change = change_handler

    # Get the next location
    result = app.loop.run_until_complete(app.location.current_location())

    # Location was returned; on_change handler was not triggered.
    assert result == toga.LatLng(16, 30)
    change_handler.assert_not_called()


def test_current_location_no_permission(app):
    """If permission has been denied, an exception is raised."""
    # Deny permission
    app.location._impl._has_permission = 0

    with pytest.raises(
        PermissionError,
        match=r"App does not have permission to use location",
    ):
        app.loop.run_until_complete(app.location.current_location())

    assert_action_performed(app.location, "has permission")
    assert_action_not_performed(app.location, "get next location")


def test_tracking(app):
    """Location tracking can be started and stopped."""
    # Set permission
    app.location._impl._has_permission = 1

    # Start location
    app.location.start_tracking()

    # Tracking has been started, but that doesn't cause a location to be found
    assert_action_performed(app.location, "has permission")
    assert_action_performed(app.location, "start location updates")
    assert_action_not_performed(app.location, "get next location")

    # Trigger a location change
    app.location._impl.simulate_update()

    assert_action_performed(app.location, "get next location")
    EventLog.reset()

    # Install an on_change handler
    change_handler = Mock()
    app.location.on_change = change_handler

    # Trigger another location change
    app.location._impl.simulate_update()

    assert_action_performed(app.location, "get next location")

    # on_change handler was triggered.
    change_handler.assert_called_once_with(
        app.location,
        location=toga.LatLng(16, 30),
        altitude=6,
    )

    EventLog.reset()

    # Stop location tracking
    app.location.stop_tracking()

    assert_action_performed(app.location, "has permission")
    assert_action_performed(app.location, "stop location updates")
    assert_action_not_performed(app.location, "get next location")


def test_tracking_no_permission(app):
    """If permission has been denied, an exception is raised by tracking commands."""
    # Deny permission
    app.location._impl._has_permission = 0

    with pytest.raises(
        PermissionError,
        match=r"App does not have permission to use location",
    ):
        app.location.start_tracking()

    assert_action_performed(app.location, "has permission")
    assert_action_not_performed(app.location, "start location updates")
    assert_action_not_performed(app.location, "get next location")
    EventLog.reset()
    with pytest.raises(
        PermissionError,
        match=r"App does not have permission to use location",
    ):
        app.location.stop_tracking()

    assert_action_performed(app.location, "has permission")
    assert_action_not_performed(app.location, "stop location updates")
    assert_action_not_performed(app.location, "get next location")



import sys

import pytest

import toga
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    assert_attribute_not_set,
)


@pytest.fixture
def progressbar():
    return toga.ProgressBar()


def test_progressbar_created(progressbar):
    """A progressbar can be created."""
    # Round trip the impl/interface
    assert progressbar._impl.interface == progressbar
    assert_action_performed(progressbar, "create ProgressBar")

    # Assert the default state of the progress bar.
    assert progressbar.max == pytest.approx(1.0)
    assert progressbar.value == pytest.approx(0.0)
    assert progressbar.is_determinate
    assert not progressbar.is_running


@pytest.mark.parametrize(
    "value, actual",
    [
        (3, 3.0),  # Integer, in range
        (7.4, 7.4),  # Float, in range
        (15, 10.0),  # Integer, above max
        (-1, 0.0),  # Integer, below min
        (12.0, 10.0),  # Float, above max
        (-42.0, 0.0),  # Float, below min
    ],
)
def test_set_value_determinate(progressbar, value, actual):
    """The value of a determinate progressbar can be set."""
    # Set the max value
    progressbar.max = 10

    # Confirm this makes the progress bar determinate
    assert progressbar.is_determinate

    # Set the value of the
    progressbar.value = value

    # Value is clipped and converted to float
    assert progressbar.value == pytest.approx(actual)


def test_set_value_indeterminate(progressbar):
    """Setting the value of an indeterminate progressbar is a no-op."""

    # Make the progressbar indeterminate
    progressbar.max = None

    # Confirm this makes the progress bar indeterminate
    assert not progressbar.is_determinate

    # Clear the event log
    EventLog.reset()

    # Set the value
    progressbar.value = 5

    # No call was made to set the value on the impl.
    assert_attribute_not_set(progressbar, "value")


@pytest.mark.parametrize(
    "value, actual, determinate",
    [
        (None, None, False),  # Non-determinate
        (42, 42.0, True),  # Integer
        (12.345, 12.345, True),  # Float
        ("37.42", 37.42, True),  # Float, but specified as a string.
    ],
)
def test_set_max(progressbar, value, actual, determinate):
    """The max value can be set."""
    progressbar.max = value

    # The maximum value has been applied, and has altered the determinate state.
    assert progressbar.max == pytest.approx(actual)
    assert progressbar.is_determinate == determinate


@pytest.mark.parametrize(
    "value, error, msg",
    [
        (
            "not a number",
            ValueError,
            r"could not convert string to float",
        ),  # String, but not a float
        (
            object(),
            TypeError,
            (
                r"must be a string or a number"
                if sys.version_info < (3, 10)
                else r"must be a string or a real number"
            ),
        ),  # Non-coercible to float
        (
            -42,
            ValueError,
            r"max value must be None, or a numerical value > 0",
        ),  # Negative integer
        (
            -1.234,
            ValueError,
            r"max value must be None, or a numerical value > 0",
        ),  # Negative float
        (
            0,
            ValueError,
            r"max value must be None, or a numerical value > 0",
        ),  # Non-positive integer
        (
            0.0,
            ValueError,
            r"max value must be None, or a numerical value > 0",
        ),  # Non-positive float
    ],
)
def test_invalid_max(progressbar, value, error, msg):
    """A max value that isn't positive raises an error."""
    with pytest.raises(error, match=msg):
        progressbar.max = value


def test_start(progressbar):
    """An activity indicator can be started."""
    # Not running initially
    assert not progressbar.is_running

    # Assert that start was not invoked on the impl as part of creation.
    assert_action_not_performed(progressbar, "start ProgressBar")

    # Start running
    progressbar.start()

    # The indicator is now running
    assert progressbar.is_running

    # The impl was triggered.
    assert_action_performed(progressbar, "start ProgressBar")


def test_already_started(progressbar):
    """If an activity indicator is already started, starting again is a no-op."""
    # Start the activity indicator
    progressbar.start()

    # Reset the event log so we can detect new events
    EventLog.reset()

    # Start the indicator again
    progressbar.start()

    # The indicator is still running
    assert progressbar.is_running

    # No action was performed.
    assert_action_not_performed(progressbar, "start ProgressBar")


def test_stop(progressbar):
    """An indicator can be stopped."""
    # Start running
    progressbar.start()

    # The indicator is running
    assert progressbar.is_running

    # Stop running
    progressbar.stop()

    # The indicator is no longer running
    assert not progressbar.is_running

    # The impl has been stopped
    assert_action_performed(progressbar, "stop ProgressBar")


def test_already_stopped(progressbar):
    """If an indicator is already stopped, stopping again is a no-op."""
    # The indicator is not running initially
    assert not progressbar.is_running

    # Stop running
    progressbar.stop()

    # The indicator is still not running
    assert not progressbar.is_running

    # No stop action was performed.
    assert_action_not_performed(progressbar, "stop ProgressBar")


def test_initially_running():
    """An activity indicator can be created in a started state."""
    # Creating a new progress bar with running=True so it is already running
    progressbar = toga.ProgressBar(running=True)

    # Indicator is running
    assert progressbar.is_running

    # Assert that start was invoked on the impl as part of creation.
    assert_action_performed(progressbar, "start ProgressBar")


def test_determinate_switch(progressbar):
    """A progressbar can switch between determinate and indeterminate."""
    # Set initial max and value
    progressbar.max = 10
    progressbar.value = 5

    # State of progress bar is determinate
    assert progressbar.is_determinate
    assert progressbar.value == pytest.approx(5.0)

    # Make the progress bar indeterminate
    progressbar.max = None

    # State of progress bar is indeterminate
    assert not progressbar.is_determinate
    assert progressbar.value is None

    # Switch back to determinate
    progressbar.max = 15

    # State of progress bar is determinate
    assert progressbar.is_determinate
    assert progressbar.value == pytest.approx(5.0)


def test_disable_no_op(progressbar):
    """ProgressBar doesn't have a disabled state."""
    # Enabled by default
    assert progressbar.enabled

    # Try to disable the widget
    progressbar.enabled = False

    # Still enabled.
    assert progressbar.enabled



import importlib
import os
import sys
from functools import lru_cache

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    # Before Python 3.10, entry_points did not support the group argument;
    # so, the backport package must be used on older versions.
    from importlib_metadata import entry_points


# Map python sys.platform with toga platforms names
_TOGA_PLATFORMS = {
    "android": "android",
    "darwin": "macOS",
    "ios": "iOS",
    "linux": "linux",
    "freebsd": "freeBSD",
    "tvos": "tvOS",
    "watchos": "watchOS",
    "wearos": "wearOS",
    "emscripten": "web",
    "win32": "windows",
}


def get_current_platform():
    # Rely on `sys.getandroidapilevel`, which only exists on Android; see
    # https://github.com/beeware/Python-Android-support/issues/8
    if hasattr(sys, "getandroidapilevel"):
        return "android"
    elif sys.platform.startswith("freebsd"):
        return "freeBSD"
    else:
        return _TOGA_PLATFORMS.get(sys.platform)


current_platform = get_current_platform()


def find_backends():
    # As of Setuptools 65.5, entry points are returned duplicated if the
    # package is installed editable. Use a set to ensure that each entry point
    # is only returned once.
    # See https://github.com/pypa/setuptools/issues/3649
    return sorted(set(entry_points(group="toga.backends")))


@lru_cache(maxsize=1)
def get_platform_factory():
    """Determine the current host platform and import the platform factory.

    If the ``TOGA_BACKEND`` environment variable is set, the factory will be loaded
    from that module.

    Raises :any:`RuntimeError` if an appropriate host platform cannot be identified.

    :returns: The factory for the host platform.
    """
    backend_value = os.environ.get("TOGA_BACKEND")
    if backend_value:
        try:
            factory = importlib.import_module(f"{backend_value}.factory")
        except ModuleNotFoundError as e:
            toga_backends_values = ", ".join(
                [f"{backend.value!r}" for backend in find_backends()]
            )
            # Android doesn't report Python exception chains in crashes
            # (https://github.com/chaquo/chaquopy/issues/890), so include the original
            # exception message in case the backend does exist but throws a
            # ModuleNotFoundError from one of its internal imports.
            raise RuntimeError(
                f"The backend specified by TOGA_BACKEND ({backend_value!r}) could "
                f"not be loaded ({e}). It should be one of: {toga_backends_values}."
            )
    else:
        toga_backends = find_backends()
        if len(toga_backends) == 0:
            raise RuntimeError("No Toga backend could be loaded.")
        elif len(toga_backends) == 1:
            backend = toga_backends[0]
        else:
            # multiple backends are installed: choose the one that matches the host platform
            matching_backends = [
                backend for backend in toga_backends if backend.name == current_platform
            ]
            if len(matching_backends) == 0:
                toga_backends_string = ", ".join(
                    [f"{backend.value!r} ({backend.name})" for backend in toga_backends]
                )
                raise RuntimeError(
                    f"Multiple Toga backends are installed ({toga_backends_string}), "
                    f"but none of them match your current platform ({current_platform!r}). "
                    "Install a backend for your current platform, or use "
                    "TOGA_BACKEND to specify a backend."
                )
            if len(matching_backends) > 1:
                toga_backends_string = ", ".join(
                    [
                        f"{backend.value!r} ({backend.name})"
                        for backend in matching_backends
                    ]
                )
                raise RuntimeError(
                    f"Multiple candidate toga backends found: ({toga_backends_string}). "
                    "Uninstall the backends you don't require, or use "
                    "TOGA_BACKEND to specify a backend."
                )
            backend = matching_backends[0]
        factory = importlib.import_module(f"{backend.value}.factory")
    return factory



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol

import toga
from toga.handlers import AsyncResult, PermissionResult, wrapped_handler
from toga.platform import get_platform_factory

if TYPE_CHECKING:
    from toga.app import App


class LocationResult(AsyncResult):
    RESULT_TYPE = "location"


class OnLocationChangeHandler(Protocol):
    def __call__(
        self,
        service: Location,
        location: toga.LatLng,
        altitude: float | None,
        **kwargs: Any,
    ) -> None:
        """A handler that will be invoked when the user's location changes.

        :param service: the location service that generated the update.
        :param location: The user's location as (latitude, longitude).
        :param altitude: The user's altitude in meters above mean sea level. Returns
            None if the altitude could not be determined.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """
        ...


class Location:
    def __init__(self, app: App):
        self.factory = get_platform_factory()
        self._app = app
        self._impl = self.factory.Location(self)

        self.on_change = None

    @property
    def app(self) -> App:
        """The app with which the location service is associated"""
        return self._app

    @property
    def has_permission(self) -> bool:
        """Does the app have permission to use location services?

        If the platform requires the user to explicitly confirm permission, and
        the user has not yet given permission, this will return ``False``.
        """
        return self._impl.has_permission()

    def request_permission(self) -> PermissionResult:
        """Request sufficient permissions to capture the user's location.

        If permission has already been granted, this will return without prompting the
        user.

        This method will only grant permission to access location services while the
        app is in the foreground. If you want your application to have permission to
        track location while the app is in the background, you must call this method,
        then make an *additional* permission request for background permissions using
        :any:`Location.request_background_permission()`.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will start the process of requesting permissions, but will return
        *immediately*. The return value can be awaited in an asynchronous context, but
        cannot be used directly.

        :returns: An asynchronous result; when awaited, returns True if the app has
            permission to capture the user's location; False otherwise.
        """
        result = PermissionResult(None)

        if has_permission := self.has_permission:
            result.set_result(has_permission)
        else:
            self._impl.request_permission(result)

        return result

    @property
    def has_background_permission(self) -> bool:
        """Does the app have permission to use location services in the background?

        If the platform requires the user to explicitly confirm permission, and the user
        has not yet given permission, this will return ``False``.
        """
        return self._impl.has_background_permission()

    def request_background_permission(self) -> PermissionResult:
        """Request sufficient permissions to capture the user's location in the
        background.

        If permission has already been granted, this will return without prompting the
        user.

        Before requesting background permission, you must first request and receive
        foreground location permission using :any:`Location.request_permission`.
        If you ask for background permission before receiving foreground location
        permission, a :any:`PermissionError` will be raised.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will start the process of requesting permissions, but will return
        *immediately*. The return value can be awaited in an asynchronous context, but
        cannot be used directly.

        :returns: An asynchronous result; when awaited, returns True if the app has
            permission to capture the user's location while running in the
            background; False otherwise.
        :raises PermissionError: If the app has not already requested and received
            permission to use location services.
        """
        result = PermissionResult(None)
        if not self.has_permission:
            result.set_exception(
                PermissionError(
                    "Cannot ask for background location permission "
                    "before confirming foreground location permission."
                )
            )
        elif has_background_permission := self.has_background_permission:
            result.set_result(has_background_permission)
        else:
            self._impl.request_background_permission(result)

        return result

    @property
    def on_change(self) -> OnLocationChangeHandler:
        """The handler to invoke when an update to the user's location is available."""
        return self._on_change

    @on_change.setter
    def on_change(self, handler):
        self._on_change = wrapped_handler(self, handler)

    def start_tracking(self):
        """Start monitoring the user's location for changes.

        An :any:`on_change` callback will be generated when the user's location
        changes.

        :raises PermissionError: If the app has not requested and received permission to
            use location services.
        """
        if self.has_permission:
            self._impl.start_tracking()
        else:
            raise PermissionError(
                "App does not have permission to use location services"
            )

    def stop_tracking(self):
        """Stop monitoring the user's location.

        :raises PermissionError: If the app has not requested and received permission to
            use location services.
        """
        if self.has_permission:
            self._impl.stop_tracking()
        else:
            raise PermissionError(
                "App does not have permission to use location services"
            )

    def current_location(self) -> LocationResult:
        """Obtain the user's current location using the location service.

        If the app hasn't requested and received permission to use location services, a
        :any:`PermissionError` will be raised.

        **This is an asynchronous method**. If you call this method in a synchronous
        context, it will start the process of requesting the user's location, but will
        return *immediately*. The return value can be awaited in an asynchronous
        context, but cannot be used directly.

        If location tracking is enabled, and an :any:`on_change` handler is installed,
        requesting the current location may also cause that handler to be invoked.

        :returns: An asynchronous result; when awaited, returns the current
            :any:`toga.LatLng` of the device.
        :raises PermissionError: If the app has not requested and received permission to
            use location services.
        """
        location = LocationResult(None)
        if self.has_permission:
            self._impl.current_location(location)
        else:
            location.set_exception(
                PermissionError("App does not have permission to use location services")
            )
        return location



from unittest.mock import Mock

from toga.sources import Row


def test_row():
    """A row can be created and modified."""
    source = Mock()
    row = Row(val1="value 1", val2=42)
    # Set a source.
    row._source = source

    # initial values are as expected
    assert row.val1 == "value 1"
    assert row.val2 == 42

    # An existing attribute can be updated.
    row.val1 = "new value"
    assert row.val1 == "new value"
    source.notify.assert_called_once_with("change", item=row)
    source.notify.reset_mock()

    # Deleting an attribute causes a change notification
    del row.val1
    assert not hasattr(row, "val1")
    source.notify.assert_called_once_with("change", item=row)
    source.notify.reset_mock()

    # Setting an attribute with an underscore isn't a notifiable event
    row._secret = "secret value"
    assert row._secret == "secret value"
    source.notify.assert_not_called()

    # An attribute that wasn't in the original attribute set
    # still causes a change notification
    row.val3 = "other value"
    assert row.val3 == "other value"
    source.notify.assert_called_once_with("change", item=row)
    source.notify.reset_mock()

    # Deleting an attribute that wasn't in the original attribute set
    # still causes a change notification
    del row.val3
    assert not hasattr(row, "val")
    source.notify.assert_called_once_with("change", item=row)
    source.notify.reset_mock()


def test_row_without_source():
    """A row with no source can be created and modified."""
    row = Row(val1="value 1", val2=42)

    # initial values are as expected
    assert row.val1 == "value 1"
    assert row.val2 == 42

    # An existing attribute can be updated.
    row.val1 = "new value"
    assert row.val1 == "new value"

    # Deleting an attribute causes a change notification
    del row.val1
    assert not hasattr(row, "val1")

    # Setting an attribute starting with an underscore isn't a notifiable event
    row._secret = "secret value"
    assert row._secret == "secret value"

    # An attribute that wasn't in the original attribute set
    # still causes a change notification
    row.val3 = "other value"
    assert row.val3 == "other value"

    # Deleting an attribute that wasn't in the original attribute set
    # still causes a change notification
    del row.val3
    assert not hasattr(row, "val")



from unittest.mock import call

from toga.colors import rgb
from toga.fonts import Font
from toga.style.pack import (
    CENTER,
    HIDDEN,
    LEFT,
    RIGHT,
    RTL,
    Pack,
)

from .utils import ExampleNode


def test_set_default_right_textalign_when_rtl():
    root = ExampleNode("app", style=Pack(text_direction=RTL))
    root.style.reapply()
    # Two calls; one caused by text_align, one because text_direction
    # implies a change to text alignment.
    assert root._impl.set_alignment.mock_calls == [call(RIGHT), call(RIGHT)]


def test_set_default_left_textalign_when_no_rtl():
    root = ExampleNode("app", style=Pack())
    root.style.reapply()
    # Two calls; one caused by text_align, one because text_direction
    # implies a change to text alignment.
    assert root._impl.set_alignment.mock_calls == [call(LEFT), call(LEFT)]


def test_set_center_alignment():
    root = ExampleNode("app", style=Pack(text_align="center"))
    root.style.reapply()
    root._impl.set_alignment.assert_called_once_with(CENTER)


def test_set_color():
    color = "#ffffff"
    root = ExampleNode("app", style=Pack(color=color))
    root.style.reapply()
    root._impl.set_color.assert_called_once_with(rgb(255, 255, 255))


def test_set_background_color():
    color = "#ffffff"
    root = ExampleNode("app", style=Pack(background_color=color))
    root.style.reapply()
    root._impl.set_background_color.assert_called_once_with(rgb(255, 255, 255))


def test_set_font():
    root = ExampleNode(
        "app",
        style=Pack(
            font_family="Roboto",
            font_size=12,
            font_style="normal",
            font_variant="small-caps",
            font_weight="bold",
        ),
    )
    root.style.reapply()
    root._impl.set_font.assert_called_with(
        Font("Roboto", 12, style="normal", variant="small-caps", weight="bold")
    )
    root.refresh.assert_called_with()


def test_set_visibility_hidden():
    root = ExampleNode("app", style=Pack(visibility=HIDDEN))
    root.style.reapply()
    root._impl.set_hidden.assert_called_once_with(True)



import asyncio
from unittest.mock import Mock

import pytest

import toga
from toga.widgets.webview import JavaScriptResult
from toga_dummy.utils import (
    assert_action_performed,
    assert_action_performed_with,
    attribute_value,
)
from toga_dummy.widgets.webview import WebView as DummyWebView


@pytest.fixture
def widget():
    return toga.WebView()


def test_widget_created():
    """A WebView can be created with minimal arguments."""
    widget = toga.WebView()

    assert widget._impl.interface == widget
    assert_action_performed(widget, "create WebView")

    assert widget.url is None
    assert widget.user_agent == "Toga dummy backend"
    assert widget._on_webview_load._raw is None


def test_create_with_values():
    """A WebView can be created with initial values."""
    on_webview_load = Mock()

    widget = toga.WebView(
        url="https://beeware.org",
        user_agent="Custom agent",
        on_webview_load=on_webview_load,
    )
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create WebView")

    assert widget.url == "https://beeware.org"
    assert widget.user_agent == "Custom agent"
    assert widget.on_webview_load._raw == on_webview_load


def test_webview_load_disabled(monkeypatch):
    """If the backend doesn't support on_webview_load, a warning is raised."""
    try:
        # Temporarily set the feature attribute on the backend
        DummyWebView.SUPPORTS_ON_WEBVIEW_LOAD = False

        # Instantiate a new widget with a hobbled backend.
        widget = toga.WebView()
        handler = Mock()

        # Setting the handler raises a warning
        with pytest.warns(
            toga.NotImplementedWarning,
            match=r"\[Dummy\] Not implemented: WebView\.on_webview_load",
        ):
            widget.on_webview_load = handler

        # But the handler is still installed
        assert widget.on_webview_load._raw == handler
    finally:
        # Clear the feature attribute.
        del DummyWebView.SUPPORTS_ON_WEBVIEW_LOAD


@pytest.mark.parametrize(
    "url",
    [
        "http://example.com",
        "https://example.com",
        None,
    ],
)
def test_url(widget, url):
    """The URL of a webview can be set."""
    # Set up a load handler
    on_webview_load_handler = Mock()
    widget.on_webview_load = on_webview_load_handler

    widget.url = url
    assert widget.url == url

    # There's no future created for the load.
    assert attribute_value(widget, "loaded_future") is None

    # The load handler hasn't been called yet
    on_webview_load_handler.assert_not_called()

    # Simulate a page load
    widget._impl.simulate_page_loaded()

    # handler has been invoked
    on_webview_load_handler.assert_called_once_with(widget)


@pytest.mark.parametrize(
    "url",
    [
        "http://example.com",
        "https://example.com",
        None,
    ],
)
async def test_load_url(widget, url):
    """The URL of a webview can be loaded asynchronously."""
    # Set up a load handler
    on_webview_load_handler = Mock()
    widget.on_webview_load = on_webview_load_handler

    # An async task that simulates a page load after a delay
    async def delayed_page_load():
        await asyncio.sleep(0.1)

        # There should be a pending future
        loaded_future = attribute_value(widget, "loaded_future")
        assert asyncio.isfuture(loaded_future)

        # Complete the page load
        widget._impl.simulate_page_loaded()

        # The loading result is None
        assert loaded_future.result() is None

    asyncio.create_task(delayed_page_load())

    # wait for a URL to load
    await widget.load_url(url)

    # The URL has been set
    assert widget.url == url

    # The handler has been invoked
    on_webview_load_handler.assert_called_once_with(widget)

    # The future has been cleared
    assert attribute_value(widget, "loaded_future") is None


@pytest.mark.parametrize(
    "url",
    [
        "file:///Path/to/file",
        "gopher://example.com",
    ],
)
async def test_invalid_url(widget, url):
    """URLs must start with https:// or http://"""
    with pytest.raises(
        ValueError,
        match=r"WebView can only display http:// and https:// URLs",
    ):
        widget.url = url

    with pytest.raises(
        ValueError,
        match=r"WebView can only display http:// and https:// URLs",
    ):
        await widget.load_url(url)


def test_set_content(widget):
    """Static HTML content can be loaded into the page."""
    widget.set_content("https://example.com", "<h1>Fancy page</h1>")
    assert_action_performed_with(
        widget,
        "set content",
        root_url="https://example.com",
        content="<h1>Fancy page</h1>",
    )


def test_user_agent(widget):
    """The user agent can be customized."""
    widget.user_agent = "New user agent"
    assert widget.user_agent == "New user agent"


def test_evaluate_javascript(widget):
    """Javascript can be evaluated."""
    result = widget.evaluate_javascript("test(1);")
    assert_action_performed(widget, "evaluate_javascript")

    assert isinstance(result, JavaScriptResult)

    # Attempting to use or compare the result raises an error
    with pytest.raises(
        RuntimeError,
        match=r"Can't check JavaScript result directly; use await or an on_result handler",
    ):
        result == 42


async def test_evaluate_javascript_async(widget):
    """Javascript can be evaluated asynchronously, and an asynchronous result
    returned."""

    # An async task that simulates evaluation of Javascript after a delay
    async def delayed_page_load():
        await asyncio.sleep(0.1)

        # Complete the Javascript
        widget._impl.simulate_javascript_result(42)

    asyncio.create_task(delayed_page_load())

    result = await widget.evaluate_javascript("test(1);")
    assert_action_performed(widget, "evaluate_javascript")

    assert result == 42


async def test_evaluate_javascript_sync(widget):
    """Deprecated sync handlers can be used for Javascript evaluation."""

    # An async task that simulates evaluation of Javascript after a delay
    async def delayed_page_load():
        await asyncio.sleep(0.1)

        # Complete the Javascript
        widget._impl.simulate_javascript_result(42)

    asyncio.create_task(delayed_page_load())

    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        result = await widget.evaluate_javascript(
            "test(1);", on_result=on_result_handler
        )

    assert_action_performed(widget, "evaluate_javascript")

    assert result == 42

    # The async handler was invoked
    on_result_handler.assert_called_once_with(42)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol

from toga.handlers import wrapped_handler
from toga.icons import Icon
from toga.keys import Key
from toga.platform import get_platform_factory

if TYPE_CHECKING:
    from toga.app import App
    from toga.icons import IconContent


class Group:
    def __init__(
        self,
        text: str,
        *,
        parent: Group | None = None,
        section: int = 0,
        order: int = 0,
    ):
        """
        A collection of commands to display together.

        :param text: A label for the group.
        :param parent: The parent of this group; use ``None`` to make a root group.
        :param section: The section where the group should appear within its parent. A
            section cannot be specified unless a parent is also specified.
        :param order: The position where the group should appear within its section.
            If multiple items have the same group, section and order, they will be
            sorted alphabetically by their text.
        """
        self.text = text
        self.order = order
        if parent is None and section != 0:
            raise ValueError("Section cannot be set without parent group")
        self.section = section

        # Prime the underlying value of _parent so that the setter has a current value
        # to work with
        self._parent = None
        self.parent = parent

    @property
    def parent(self) -> Group | None:
        """The parent of this group; returns ``None`` if the group is a root group."""
        return self._parent

    @parent.setter
    def parent(self, parent: Group | None):
        if parent is None:
            self._parent = None
        elif parent == self:
            raise ValueError("A group cannot be it's own parent")
        elif self.is_parent_of(parent):
            raise ValueError(
                f"Cannot set parent; {self.text!r} is an ancestor of {parent.text!r}."
            )
        else:
            self._parent = parent

    @property
    def root(self) -> Group:
        """The root group for this group.

        This will be ``self`` if the group *is* a root group."""
        if self.parent is None:
            return self
        return self.parent.root

    def is_parent_of(self, child: Group | None) -> bool:
        """Is this group a parent of the provided group, directly or indirectly?

        :param child: The potential child to check
        :returns: True if this group is a parent of the provided child.
        """
        if child is None:
            return False
        if child.parent is None:
            return False
        if child.parent == self:
            return True
        return self.is_parent_of(child.parent)

    def is_child_of(self, parent: Group | None) -> bool:
        """Is this group a child of the provided group, directly or indirectly?

        :param parent: The potential parent to check
        :returns: True if this group is a child of the provided parent.
        """
        if parent is None:
            return False
        return parent.is_parent_of(self)

    def __hash__(self) -> int:
        return hash(self.key)

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, (Group, Command)):
            return False
        return self.key < other.key

    def __gt__(self, other: Any) -> bool:
        if not isinstance(other, (Group, Command)):
            return False
        return other < self

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, (Group, Command)):
            return False
        return self.key == other.key

    def __repr__(self) -> str:
        parent_string = (
            f" parent={self.parent} section={self.section}"
            if self.parent is not None
            else ""
        )
        return f"<Group text={self.text!r} order={self.order}{parent_string}>"

    @property
    def key(self) -> tuple[(int, int, str)]:
        """A unique tuple describing the path to this group."""
        self_tuple = (self.section, self.order, self.text)
        if self.parent is None:
            return tuple([self_tuple])
        return tuple([*self.parent.key, self_tuple])

    # Standard groups - docstrings can only be provided within the `class` statement,
    # but the objects can't be instantiated here.
    APP = None  #: Application-level commands
    FILE = None  #: File commands
    EDIT = None  #: Editing commands
    VIEW = None  #: Content appearance commands
    COMMANDS = None  #: Default group for user-provided commands
    WINDOW = None  #: Window management commands
    HELP = None  #: Help commands


Group.APP = Group("*", order=0)
Group.FILE = Group("File", order=1)
Group.EDIT = Group("Edit", order=10)
Group.VIEW = Group("View", order=20)
Group.COMMANDS = Group("Commands", order=30)
Group.WINDOW = Group("Window", order=90)
Group.HELP = Group("Help", order=100)


class ActionHandler(Protocol):
    def __call__(self, command: Command, **kwargs) -> bool:
        """A handler that will be invoked when a Command is invoked.

        :param command: The command that triggered the action.
        :param kwargs: Ensures compatibility with additional arguments introduced in
            future versions.
        """
        ...


class Command:
    def __init__(
        self,
        action: ActionHandler | None,
        text: str,
        *,
        shortcut: str | Key | None = None,
        tooltip: str | None = None,
        icon: IconContent | None = None,
        group: Group = Group.COMMANDS,
        section: int = 0,
        order: int = 0,
        enabled: bool = True,
    ):
        """
        Create a new Command.

        Commands may not use all the arguments - for example, on some platforms, menus
        will contain icons; on other platforms they won't.

        :param action: A handler to invoke when the command is activated. If this is
            ``None``, the command will be disabled.
        :param text: A label for the command.
        :param shortcut: A key combination that can be used to invoke the command.
        :param tooltip: A short description of what the command will do.
        :param icon: The :any:`icon content <IconContent>` that can be used to decorate
            the command if the platform requires.
        :param group: The group to which this command belongs.
        :param section: The section where the command should appear within its group.
        :param order: The position where the command should appear within its section.
            If multiple items have the same group, section and order, they will be
            sorted alphabetically by their text.
        :param enabled: Is the Command currently enabled?
        """
        self.text = text

        self.shortcut = shortcut
        self.tooltip = tooltip
        self.icon = icon

        self.group = group
        self.section = section
        self.order = order

        self.action = action

        self.factory = get_platform_factory()
        self._impl = self.factory.Command(interface=self)

        self.enabled = enabled

    @property
    def key(self) -> tuple[(int, int, str)]:
        """A unique tuple describing the path to this command.

        Each element in the tuple describes the (section, order, text) for the
        groups that must be navigated to invoke this action.
        """
        return tuple([*self.group.key, (self.section, self.order, self.text)])

    @property
    def enabled(self) -> bool:
        """Is the command currently enabled?"""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = value and getattr(self.action, "_raw", True) is not None
        self._impl.set_enabled(value)

    @property
    def icon(self) -> Icon | None:
        """The Icon for the command.

        Can be specified as any valid :any:`icon content <IconContent>`.
        """
        return self._icon

    @icon.setter
    def icon(self, icon_or_name: IconContent | None):
        if isinstance(icon_or_name, Icon) or icon_or_name is None:
            self._icon = icon_or_name
        else:
            self._icon = Icon(icon_or_name)

    @property
    def action(self) -> ActionHandler | None:
        """The Action attached to the command."""
        return self._action

    @action.setter
    def action(self, action: ActionHandler | None):
        """Set the action attached to the command

        Needs to be a valid ActionHandler or ``None``
        """
        self._action = wrapped_handler(self, action)

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, (Group, Command)):
            return False
        return self.key < other.key

    def __gt__(self, other: Any) -> bool:
        if not isinstance(other, (Group, Command)):
            return False
        return other < self

    def __repr__(self) -> bool:
        return (
            f"<Command text={self.text!r} "
            f"group={self.group} "
            f"section={self.section} "
            f"order={self.order}>"
        )


class Separator:
    def __init__(self, group: Group = None):
        """A representation of a separator between sections in a Group.

        :param group: The group that contains the separator.
        """
        self.group = group

    def __repr__(self) -> str:
        return f"<Separator group={self.group.text}>"

    def __eq__(self, other) -> bool:
        if isinstance(other, Separator):
            return self.group == other.group
        return False


class CommandSetChangeHandler(Protocol):
    def __call__(self) -> None:
        """A handler that will be invoked when a Command or Group is added to the CommandSet."""
        ...


class CommandSet:
    def __init__(
        self,
        on_change: CommandSetChangeHandler = None,
        app: App | None = None,
    ):
        """
        A collection of commands.

        This is used as an internal representation of Menus, Toolbars, and any other
        graphical manifestations of commands. You generally don't need to construct a
        CommandSet of your own; you should use existing app or window level CommandSet
        instances.

        The collection can be iterated over to provide the display order of the commands
        managed by the group.

        :param on_change: A method that should be invoked when this command set changes.
        :param app: The app this command set is associated with, if it is not the app's
            own commandset.
        """
        self._app = app
        self._commands = set()
        self.on_change = on_change

    def add(self, *commands: Command | Group):
        if self.app and self.app is not None:
            self.app.commands.add(*commands)
        self._commands.update(commands)
        if self.on_change:
            self.on_change()

    def clear(self):
        self._commands = set()
        if self.on_change:
            self.on_change()

    @property
    def app(self) -> App:
        return self._app

    def __len__(self) -> int:
        return len(self._commands)

    def __iter__(self) -> Command | Separator:
        cmd_iter = iter(sorted(self._commands))

        def descendant(group, ancestor):
            # Return the immediate descendant of ancestor used by this group.
            if group.parent == ancestor:
                return group
            if group.parent:
                return descendant(group.parent, ancestor)
            return None

        # The iteration over commands tells us the exact order of commands, but doesn't
        # tell us anything about menu and submenu structure. In order to insert section
        # breaks in the right place (including before and after submenus), we need to
        # iterate over commands inside each group, dealing with each subgroup as an
        # independent iteration.
        #
        # The iterator over commands is maintained external to this recursive iteration,
        # because we may need to inspect the command at multiple group levels, but we
        # can't `peek` at the top element of an iterator, `push` an item back on after
        # it has been consumed, or pass the consumed item as a return value in addition
        # to the generator result.
        def _iter_group(parent):
            nonlocal command
            nonlocal finished
            section = None

            def _section_break(obj):
                # Utility method that will insert a section break, if required.
                # A section break is needed if the section for the object we're
                # processing (either a command, or a group acting as a submenu)
                # has a section ID different to the previous object processed at
                # this level, excluding the very first object (as there's no need
                # for a section break before the first command/submenu).
                nonlocal section
                if section is not None:
                    if section != obj.section:
                        yield Separator(parent)
                        section = obj.section
                else:
                    section = obj.section

            while not finished:
                if parent is None:
                    # Handle root-level menus
                    yield from _iter_group(command.group.root)
                elif command.group == parent:
                    # A normal command at this level of the group.
                    yield from _section_break(command)
                    yield command

                    # Consume the next item on the iterator; if we run out, mark the
                    # sentinel that says we've finished. We can't just raise
                    # StopIteration, because that stops the *generator* we're creating.
                    try:
                        command = next(cmd_iter)
                    except StopIteration:
                        finished = True
                else:
                    # The command isn't in this group. If the command is in descendant
                    # group, yield items from the group. If it's not a descendant, then
                    # there are no more commands in this group; we can return to the
                    # previous group for processing.
                    subgroup = descendant(command.group, parent)
                    if subgroup:
                        yield from _section_break(subgroup)
                        yield from _iter_group(subgroup)
                    else:
                        return

        # Prime the initial command into the command iterator
        try:
            command = next(cmd_iter)
        except StopIteration:
            pass
        else:
            finished = False
            yield from _iter_group(None)



from __future__ import annotations

import warnings
from typing import Any

from toga.handlers import wrapped_handler
from toga.sources import Node, Source, TreeSource
from toga.sources.accessors import build_accessors, to_accessor

from .base import Widget


class Tree(Widget):
    def __init__(
        self,
        headings: list[str] | None = None,
        id=None,
        style=None,
        data: Any = None,
        accessors: list[str] | None = None,
        multiple_select: bool = False,
        on_select: callable | None = None,
        on_activate: callable | None = None,
        missing_value: str = "",
        on_double_click=None,  # DEPRECATED
    ):
        """Create a new Tree widget.

        :param headings: The column headings for the tree. Headings can only contain one
            line; any text after a newline will be ignored.

            A value of :any:`None` will produce a table without headings.
            However, if you do this, you *must* give a list of accessors.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param data: Initial :any:`data` to be displayed in the tree.

        :param accessors: Defines the attributes of the data source that will be used to
            populate each column. Must be either:

            * ``None`` to derive accessors from the headings, as described above; or
            * A list of the same size as ``headings``, specifying the accessors for each
              heading. A value of :any:`None` will fall back to the default generated
              accessor; or
            * A dictionary mapping headings to accessors. Any missing headings will fall
              back to the default generated accessor.

        :param multiple_select: Does the tree allow multiple selection?
        :param on_select: Initial :any:`on_select` handler.
        :param on_activate: Initial :any:`on_activate` handler.
        :param missing_value: The string that will be used to populate a cell when the
            value provided by its accessor is :any:`None`, or the accessor isn't
            defined.
        :param on_double_click: **DEPRECATED**; use :attr:`on_activate`.
        """
        super().__init__(id=id, style=style)

        ######################################################################
        # 2023-06: Backwards compatibility
        ######################################################################
        if on_double_click:
            if on_activate:
                raise ValueError("Cannot specify both on_double_click and on_activate")
            else:
                warnings.warn(
                    "Tree.on_double_click has been renamed Tree.on_activate.",
                    DeprecationWarning,
                )
                on_activate = on_double_click
        ######################################################################
        # End backwards compatibility.
        ######################################################################

        if headings is not None:
            self._headings = [heading.split("\n")[0] for heading in headings]
            self._accessors = build_accessors(self._headings, accessors)
        elif accessors is not None:
            self._headings = None
            self._accessors = accessors
        else:
            raise ValueError(
                "Cannot create a tree without either headings or accessors"
            )
        self._multiple_select = multiple_select
        self._missing_value = missing_value or ""

        # Prime some properties that need to exist before the tree is created.
        self.on_select = None
        self.on_activate = None
        self._data = None

        self._impl = self.factory.Tree(interface=self)
        self.data = data

        self.on_select = on_select
        self.on_activate = on_activate

    @property
    def enabled(self) -> bool:
        """Is the widget currently enabled? i.e., can the user interact with the widget?
        Tree widgets cannot be disabled; this property will always return True; any
        attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value):
        pass

    def focus(self):
        "No-op; Tree cannot accept input focus"
        pass

    @property
    def data(self) -> TreeSource:
        """The data to display in the tree.

        When setting this property:

        * A :any:`Source` will be used as-is. It must either be a :any:`TreeSource`, or
          a custom class that provides the same methods.

        * A value of None is turned into an empty TreeSource.

        * Otherwise, the value must be a dictionary or an iterable, which is copied
          into a new TreeSource as shown :ref:`here <treesource-item>`.
        """
        return self._data

    @data.setter
    def data(self, data: Any):
        if data is None:
            self._data = TreeSource(accessors=self._accessors, data=[])
        elif isinstance(data, Source):
            self._data = data
        else:
            self._data = TreeSource(accessors=self._accessors, data=data)

        self._data.add_listener(self._impl)
        self._impl.change_source(source=self._data)

    @property
    def multiple_select(self) -> bool:
        """Does the tree allow multiple rows to be selected?"""
        return self._multiple_select

    @property
    def selection(self) -> list[Node] | Node | None:
        """The current selection of the tree.

        If multiple selection is enabled, returns a list of Node objects from the data
        source matching the current selection. An empty list is returned if no nodes are
        selected.

        If multiple selection is *not* enabled, returns the selected Node object, or
        :any:`None` if no node is currently selected.
        """
        return self._impl.get_selection()

    def expand(self, node: Node | None = None):
        """Expand the specified node of the tree.

        If no node is provided, all nodes of the tree will be expanded.

        If the provided node is a leaf node, or the node is already expanded, this is a
        no-op.

        If a node is specified, the children of that node will also be expanded.

        :param node: The node to expand
        """
        if node is None:
            self._impl.expand_all()
        else:
            self._impl.expand_node(node)

    def collapse(self, node: Node | None = None):
        """Collapse the specified node of the tree.

        If no node is provided, all nodes of the tree will be collapsed.

        If the provided node is a leaf node, or the node is already collapsed,
        this is a no-op.

        :param node: The node to collapse
        """
        if node is None:
            self._impl.collapse_all()
        else:
            self._impl.collapse_node(node)

    def append_column(self, heading: str, accessor: str | None = None):
        """Append a column to the end of the tree.

        :param heading: The heading for the new column.
        :param accessor: The accessor to use on the data source when populating
            the tree. If not specified, an accessor will be derived from the
            heading.
        """
        self.insert_column(len(self._accessors), heading, accessor=accessor)

    def insert_column(
        self,
        index: int | str,
        heading: str | None,
        accessor: str | None = None,
    ):
        """Insert an additional column into the tree.

        :param index: The index at which to insert the column, or the accessor of the
            column before which the column should be inserted.
        :param heading: The heading for the new column. If the tree doesn't have
            headings, the value will be ignored.
        :param accessor: The accessor to use on the data source when populating the
            tree. If not specified, an accessor will be derived from the heading. An
            accessor *must* be specified if the tree doesn't have headings.
        """
        if self._headings is None:
            if accessor is None:
                raise ValueError("Must specify an accessor on a tree without headings")
            heading = None
        elif not accessor:
            accessor = to_accessor(heading)

        if isinstance(index, str):
            index = self._accessors.index(index)
        else:
            # Re-interpret negative indices, and clip indices outside valid range.
            if index < 0:
                index = max(len(self._accessors) + index, 0)
            else:
                index = min(len(self._accessors), index)

        if self._headings is not None:
            self._headings.insert(index, heading)
        self._accessors.insert(index, accessor)

        self._impl.insert_column(index, heading, accessor)

    def remove_column(self, column: int | str):
        """Remove a tree column.

        :param column: The index of the column to remove, or the accessor of the column
            to remove.
        """
        if isinstance(column, str):
            # Column is a string; use as-is
            index = self._accessors.index(column)
        else:
            if column < 0:
                index = len(self._accessors) + column
            else:
                index = column

        # Remove column
        if self._headings is not None:
            del self._headings[index]
        del self._accessors[index]
        self._impl.remove_column(index)

    @property
    def headings(self) -> list[str]:
        """The column headings for the tree (read-only)"""
        return self._headings

    @property
    def accessors(self) -> list[str]:
        """The accessors used to populate the tree (read-only)"""
        return self._accessors

    @property
    def missing_value(self) -> str:
        """The value that will be used when a data row doesn't provide a value for an
        attribute.
        """
        return self._missing_value

    @property
    def on_select(self) -> callable:
        """The callback function that is invoked when a row of the tree is selected."""
        return self._on_select

    @on_select.setter
    def on_select(self, handler: callable):
        self._on_select = wrapped_handler(self, handler)

    @property
    def on_activate(self) -> callable:
        """The callback function that is invoked when a row of the tree is activated,
        usually with a double click or similar action."""
        return self._on_activate

    @on_activate.setter
    def on_activate(self, handler):
        self._on_activate = wrapped_handler(self, handler)

    ######################################################################
    # 2023-06: Backwards compatibility
    ######################################################################

    @property
    def on_double_click(self):
        """**DEPRECATED**: Use ``on_activate``"""
        warnings.warn(
            "Tree.on_double_click has been renamed Tree.on_activate.",
            DeprecationWarning,
        )
        return self.on_activate

    @on_double_click.setter
    def on_double_click(self, handler):
        warnings.warn(
            "Tree.on_double_click has been renamed Tree.on_activate.",
            DeprecationWarning,
        )
        self.on_activate = handler



from unittest.mock import Mock

import pytest

import toga
from toga.sources import ListSource
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def on_select_handler():
    return Mock()


@pytest.fixture
def on_refresh_handler():
    return Mock(return_value=None)


@pytest.fixture
def on_primary_action_handler():
    return Mock()


@pytest.fixture
def on_secondary_action_handler():
    return Mock()


@pytest.fixture
def source():
    return ListSource(
        accessors=["key", "value", "icon"],
        data=[
            {"key": "first", "value": 111, "other": "aaa"},
            {"key": "second", "value": 222, "other": "bbb"},
            {"key": "third", "value": 333, "other": "ccc"},
        ],
    )


@pytest.fixture
def detailedlist(
    source,
    on_select_handler,
    on_refresh_handler,
    on_primary_action_handler,
    on_secondary_action_handler,
):
    return toga.DetailedList(
        accessors=["key", "value", "icon"],
        data=source,
        on_select=on_select_handler,
        on_refresh=on_refresh_handler,
        on_primary_action=on_primary_action_handler,
        on_secondary_action=on_secondary_action_handler,
    )


def test_detailedlist_created():
    """A minimal DetailedList can be created."""
    detailedlist = toga.DetailedList()
    assert detailedlist._impl.interface == detailedlist
    assert_action_performed(detailedlist, "create DetailedList")

    assert len(detailedlist.data) == 0
    assert detailedlist.accessors == ("title", "subtitle", "icon")
    assert detailedlist.missing_value == ""
    assert detailedlist.on_select._raw is None
    assert detailedlist.on_refresh._raw is None
    assert detailedlist.on_primary_action._raw is None
    assert detailedlist.on_secondary_action._raw is None
    assert detailedlist._primary_action == "Delete"
    assert detailedlist._secondary_action == "Action"

    assert_action_performed_with(detailedlist, "refresh enabled", enabled=False)
    assert_action_performed_with(detailedlist, "primary action enabled", enabled=False)
    assert_action_performed_with(
        detailedlist, "secondary action enabled", enabled=False
    )


def test_create_with_values(
    source,
    on_select_handler,
    on_refresh_handler,
    on_primary_action_handler,
    on_secondary_action_handler,
):
    """A DetailedList can be created with initial values."""
    detailedlist = toga.DetailedList(
        data=source,
        accessors=("key", "value", "icon"),
        missing_value="Boo!",
        on_select=on_select_handler,
        on_refresh=on_refresh_handler,
        primary_action="Primary",
        on_primary_action=on_primary_action_handler,
        secondary_action="Secondary",
        on_secondary_action=on_secondary_action_handler,
    )
    assert detailedlist._impl.interface == detailedlist
    assert_action_performed(detailedlist, "create DetailedList")

    assert len(detailedlist.data) == 3
    assert detailedlist.accessors == ("key", "value", "icon")
    assert detailedlist.missing_value == "Boo!"
    assert detailedlist.on_select._raw == on_select_handler
    assert detailedlist.on_refresh._raw == on_refresh_handler
    assert detailedlist.on_primary_action._raw == on_primary_action_handler
    assert detailedlist.on_secondary_action._raw == on_secondary_action_handler
    assert detailedlist._primary_action == "Primary"
    assert detailedlist._secondary_action == "Secondary"

    assert_action_performed_with(detailedlist, "refresh enabled", enabled=True)
    assert_action_performed_with(detailedlist, "primary action enabled", enabled=True)
    assert_action_performed_with(detailedlist, "secondary action enabled", enabled=True)


def test_disable_no_op(detailedlist):
    """DetailedList doesn't have a disabled state."""
    # Enabled by default
    assert detailedlist.enabled

    # Try to disable the widget
    detailedlist.enabled = False

    # Still enabled.
    assert detailedlist.enabled


def test_focus_noop(detailedlist):
    """Focus is a no-op."""

    detailedlist.focus()
    assert_action_not_performed(detailedlist, "focus")


@pytest.mark.parametrize(
    "data, all_attributes, extra_attributes",
    [
        # List of lists
        (
            [
                ["Alice", 123, "icon1"],
                ["Bob", 234, "icon2"],
                ["Charlie", 345, "icon3"],
            ],
            True,
            False,
        ),
        # List of tuples
        (
            [
                ("Alice", 123, "icon1"),
                ("Bob", 234, "icon2"),
                ("Charlie", 345, "icon3"),
            ],
            True,
            False,
        ),
        # List of dictionaries
        (
            [
                {"key": "Alice", "value": 123, "icon": "icon1", "extra": "extra1"},
                {"key": "Bob", "value": 234, "icon": "icon2", "extra": "extra2"},
                {"key": "Charlie", "value": 345, "icon": "icon3", "extra": "extra3"},
            ],
            True,
            True,
        ),
        # List of bare data
        (
            [
                "Alice",
                1234,
                "Charlie",
            ],
            False,
            False,
        ),
    ],
)
def test_set_data(
    detailedlist,
    on_select_handler,
    data,
    all_attributes,
    extra_attributes,
):
    """Data can be set from a variety of sources."""

    # The selection hasn't changed yet.
    on_select_handler.assert_not_called()

    # Change the data
    detailedlist.data = data

    # This triggered the select handler
    on_select_handler.assert_called_once_with(detailedlist)

    # A ListSource has been constructed
    assert isinstance(detailedlist.data, ListSource)
    assert len(detailedlist.data) == 3

    # The accessors are mapped in order.
    assert detailedlist.data[0].key == "Alice"
    assert detailedlist.data[2].key == "Charlie"

    if all_attributes:
        assert detailedlist.data[1].key == "Bob"

        assert detailedlist.data[0].value == 123
        assert detailedlist.data[1].value == 234
        assert detailedlist.data[2].value == 345

        assert detailedlist.data[0].icon == "icon1"
        assert detailedlist.data[1].icon == "icon2"
        assert detailedlist.data[2].icon == "icon3"
    else:
        assert detailedlist.data[1].key == 1234

    if extra_attributes:
        assert detailedlist.data[0].extra == "extra1"
        assert detailedlist.data[1].extra == "extra2"
        assert detailedlist.data[2].extra == "extra3"


def test_selection(detailedlist, on_select_handler):
    """The current selection can be retrieved."""
    # Selection is initially empty
    assert detailedlist.selection is None
    on_select_handler.assert_not_called()

    # Select an item
    detailedlist._impl.simulate_selection(1)

    # Selection returns a single row
    assert detailedlist.selection == detailedlist.data[1]

    # Selection handler was triggered
    on_select_handler.assert_called_once_with(detailedlist)


def test_refresh(detailedlist, on_refresh_handler):
    """Completion of a refresh event triggers the cleanup handler."""
    # Stimulate a refresh.
    detailedlist._impl.stimulate_refresh()

    # refresh handler was invoked
    on_refresh_handler.assert_called_once_with(detailedlist)

    # The post-refresh handler was invoked on the backend
    assert_action_performed_with(
        detailedlist,
        "after on refresh",
        widget=detailedlist,
        result=None,
    )


def test_scroll_to_top(detailedlist):
    """A DetailedList can be scrolled to the top."""
    detailedlist.scroll_to_top()

    assert_action_performed_with(detailedlist, "scroll to row", row=0)


@pytest.mark.parametrize(
    "row, effective",
    [
        # Positive index
        (0, 0),
        (2, 2),
        # Greater index than available rows
        (10, 3),
        # Negative index
        (-1, 2),
        (-3, 0),
        # Greater negative index than available rows
        (-10, 0),
    ],
)
def test_scroll_to_row(detailedlist, row, effective):
    """A DetailedList can be scrolled to a specific row."""
    detailedlist.scroll_to_row(row)

    assert_action_performed_with(detailedlist, "scroll to row", row=effective)


def test_scroll_to_row_no_data(detailedlist):
    """If there's no data, scrolling is a no-op."""
    detailedlist.data.clear()

    detailedlist.scroll_to_row(5)

    assert_action_not_performed(detailedlist, "scroll to row")


def test_scroll_to_bottom(detailedlist):
    """A DetailedList can be scrolled to the top."""
    detailedlist.scroll_to_bottom()

    assert_action_performed_with(detailedlist, "scroll to row", row=2)


######################################################################
# 2023-07: Backwards compatibility
######################################################################
def test_deprecated_names(on_primary_action_handler):
    """Deprecated names still work."""

    # Can't specify both on_delete and on_primary_action
    with pytest.raises(
        ValueError,
        match=r"Cannot specify both on_delete and on_primary_action",
    ):
        toga.DetailedList(on_delete=Mock(), on_primary_action=Mock())

    # on_delete is redirected at construction
    with pytest.warns(
        DeprecationWarning,
        match="DetailedList.on_delete has been renamed DetailedList.on_primary_action",
    ):
        select = toga.DetailedList(on_delete=on_primary_action_handler)

    # on_delete accessor is redirected to on_primary_action
    with pytest.warns(
        DeprecationWarning,
        match="DetailedList.on_delete has been renamed DetailedList.on_primary_action",
    ):
        assert select.on_delete._raw == on_primary_action_handler

    assert select.on_primary_action._raw == on_primary_action_handler

    # on_delete mutator is redirected to on_primary_action
    new_handler = Mock()
    with pytest.warns(
        DeprecationWarning,
        match="DetailedList.on_delete has been renamed DetailedList.on_primary_action",
    ):
        select.on_delete = new_handler

    assert select.on_primary_action._raw == new_handler



import sys
import types

# Before we import toga, check for the --interactive flag
# If that flag exists, mock the behavior of an interactive shell
# - replace the __main__ module with an in-memory representation.
sys.modules["__main__"] = types.ModuleType("__main__")

import toga  # noqa: E402


class SubclassedApp(toga.App):
    pass


def main():
    app = SubclassedApp("Interactive App", "org.testbed.interactive-app")

    print(f"app.paths.app={app.paths.app.resolve()}")
    print(f"app.paths.config={app.paths.config.resolve()}")
    print(f"app.paths.data={app.paths.data.resolve()}")
    print(f"app.paths.cache={app.paths.cache.resolve()}")
    print(f"app.paths.logs={app.paths.logs.resolve()}")
    print(f"app.paths.toga={app.paths.toga.resolve()}")


if __name__ == "__main__":
    main()
